from __future__ import annotations
import typing as ty
from pathlib import Path, PurePath
import json
import pkg_resources
import logging
from urllib.parse import urlparse
import site
import attrs
from arcana.core import PACKAGE_NAME
from arcana.core.exceptions import ArcanaBuildError
from arcana.core.utils.serialize import ObjectListConverter

logger = logging.getLogger("arcana")


@attrs.define
class ContainerAuthor:

    name: str
    email: str


@attrs.define
class KnownIssue:

    url: str


@attrs.define(kw_only=True)
class BaseImage:

    name: str = attrs.field(default="ubuntu")
    tag: str = attrs.field()
    package_manager: str = attrs.field()

    DEFAULT_UBUNTU_TAG = "kinetic"  # FIXME: Should revert to latest LTS "jammy"

    @property
    def reference(self):
        if self.tag:
            reference = f"{self.name}:{self.tag}"
        else:
            reference = self.name
        return reference

    @name.validator
    def name_validator(self, _, name):
        if name == "alpine":
            raise ValueError(
                "Neurodocker (the package used to build the images) does not currently "
                "support alpine base images"
            )

    @tag.default
    def tag_default(self):
        if self.name == "ubuntu":
            tag = self.DEFAULT_UBUNTU_TAG
        else:
            tag = None
        return tag

    @package_manager.default
    def package_manager_default(self):
        if self.name in ("ubuntu", "debian"):
            package_manager = "apt"
        elif self.name in ("fedora", "centos"):
            package_manager = "yum"
        else:
            package_manager = None
        return package_manager

    @package_manager.validator
    def package_manager_validator(self, _, package_manager):
        if package_manager is None:
            raise ValueError(
                f"Package manager must be supplied explicitly for unknown base image "
                f"'{self.name}' (note only 'apt' and 'yum' package managers are "
                "currently supported)"
            )
        if package_manager not in ("yum", "apt"):
            raise ValueError(
                f"Unsupported package manager '{package_manager}' provided. Only 'apt' "
                "and 'yum' package managers are currently supported by Neurodocker"
            )


@attrs.define
class License:
    """Specification of a software license that needs to be present in the container
    when the command is run.

    Parameters
    ----------
    name : str
        a name to refer to the license with. Must be unique among the licenses used
        pipelines applied to a dataset, ideally for a site. Typically named closely
        after the package it is used for along with a version number if the license,
        needs to be updated with new versions e.g. freesurfer, fsl, matlab_v2022a etc...
    destination : PurePath
        destination within the container to install the license
    description : str
        a short description of the license and what it is used for
    info_url : str
        link to website with information about license, particularly how to download it
    source : Path, optional
        path to the location of a valid license file
    store_in_image : bool
        whether the license can be stored in the image or not according to the license
        conditions
    """

    name: str = attrs.field()
    destination: PurePath = attrs.field(converter=PurePath)
    description: str = attrs.field()
    info_url: str = attrs.field()
    source: Path = attrs.field(
        default=None, converter=lambda x: Path(x) if x is not None else None
    )
    store_in_image: bool = False

    @info_url.validator
    def info_url_validator(self, _, info_url):
        parsed = urlparse(info_url)
        if not parsed.scheme or not parsed.netloc:
            raise ValueError(
                f"Could not parse info url '{info_url}', please include URL scheme"
            )

    # FIXME: this doesn't work inside images
    # @source.validator
    # def source_validator(self, _, source):
    #     if source is not None and not source.exists():
    #         raise ValueError(
    #             f"Source file for {self.name} license, '{str(source)}', does not exist"
    #         )

    @classmethod
    def column_path(self, name):
        """The column name (and resource name) for the license if it is to be downloaded
        from the source dataset"""
        return name + self.COLUMN_SUFFIX

    COLUMN_SUFFIX = "_LICENSE"


@attrs.define
class BasePackage:

    name: str
    version: str = attrs.field(
        default=None, converter=lambda v: str(v) if v is not None else None
    )


@attrs.define
class PipPackage(BasePackage):
    """Specification of a Python package"""

    url: str = None
    file_path: str = None
    extras: ty.List[str] = attrs.field(factory=list)

    @classmethod
    def unique(cls, pip_specs: ty.Iterable, remove_arcana: bool = False):
        """Merge a list of Pip install specs so each package only appears once

        Parameters
        ----------
        pip_specs : ty.Iterable[PipPackage]
            the pip specs to merge
        remove_arcana : bool
            remove arcana if present from the merged list

        Returns
        -------
        list[PipPackage]
            the merged pip specs

        Raises
        ------
        ArcanaError
            if there is a mismatch between two entries of the same package
        """
        dct = {}
        for pip_spec in pip_specs:
            if isinstance(pip_spec, dict):
                pip_spec = PipPackage(**pip_spec)
            if pip_spec.name == PACKAGE_NAME and remove_arcana:
                continue
            try:
                prev_spec = dct[pip_spec.name]
            except KeyError:
                dct[pip_spec.name] = pip_spec
            else:
                if (
                    prev_spec.version != pip_spec.version
                    or prev_spec.url != pip_spec.url
                    or prev_spec.file_path != pip_spec.file_path
                ):
                    raise RuntimeError(
                        f"Cannot install '{pip_spec.name}' due to conflict "
                        f"between requested versions, {pip_spec} and {prev_spec}"
                    )
                prev_spec.extras.extend(pip_spec.extras)
        return list(dct.values())

    def local_package_location(self, pypi_fallback: bool = False):
        """Detect the installed locations of the packages, including development
        versions.

        Parameters
        ----------
        package : [PipPackage]
            the packages (or names of) the versions to detect
        pypi_fallback : bool, optional
            Fallback to PyPI version if requested version isn't installed locally

        Returns
        -------
        PipPackage
            the pip specification for the installation location of the package
        """
        try:
            pkg = next(
                p for p in pkg_resources.working_set if p.project_name == self.name
            )
        except StopIteration:
            if pypi_fallback:
                logger.info(
                    f"Did not find local installation of package {self.name} "
                    "falling back to installation from PyPI"
                )
                return self
            raise ArcanaBuildError(
                f"Did not find {self.name} in installed working set:\n"
                + "\n".join(
                    sorted(
                        p.key + "/" + p.project_name for p in pkg_resources.working_set
                    )
                )
            )
        if (
            self.version
            and (
                not (pkg.version.endswith(".dirty") or self.version.endswith(".dirty"))
            )
            and pkg.version != self.version
        ):
            msg = (
                f"Requested package {self.name}=={self.version} does "
                "not match installed " + pkg.version
            )
            if pypi_fallback:
                logger.warning(msg + " falling back to installation from PyPI")
                return self
            raise ArcanaBuildError(msg)
        pkg_loc = Path(pkg.location).resolve()
        # Determine whether installed version of requirement is locally
        # installed (and therefore needs to be copied into image) or can
        # be just downloaded from PyPI
        if pkg_loc not in site_pkg_locs:
            # Copy package into Docker image and instruct pip to install from
            # that copy
            local_spec = PipPackage(
                name=self.name, file_path=pkg_loc, extras=self.extras
            )
        else:
            # Check to see whether package is installed via "direct URL" instead
            # of through PyPI
            direct_url_path = Path(pkg.egg_info) / "direct_url.json"
            if direct_url_path.exists():
                with open(direct_url_path) as f:
                    url_spec = json.load(f)
                url = url_spec["url"]
                vcs_info = url_spec.get(
                    "vcs_info", url_spec
                )  # Fallback to trying to find VCS info in the base url-spec dict
                if url.startswith("file://"):
                    local_spec = PipPackage(
                        name=self.name,
                        file_path=url[len("file://") :],
                        extras=self.extras,
                    )
                else:
                    vcs_info = url_spec.get("vcs_info", url_spec)
                    if "vcs" in vcs_info:
                        url = vcs_info["vcs"] + "+" + url
                    if "commit_id" in vcs_info:
                        url += "@" + vcs_info["commit_id"]
                    local_spec = PipPackage(name=self.name, url=url, extras=self.extras)
            else:
                local_spec = PipPackage(
                    name=self.name, version=pkg.version, extras=self.extras
                )
        return local_spec


@attrs.define
class SystemPackage(BasePackage):

    pass


@attrs.define
class CondaPackage(BasePackage):

    REQUIRED = ["numpy", "traits"]  # FIXME: Not sure if traits is actually required


@attrs.define
class NeurodockerTemplate:

    name: str
    version: str


def python_package_converter(packages):
    """
    Split out and merge any extras specifications (e.g. "arcana[test]")
    between dependencies of the same package
    """
    return PipPackage.unique(
        ObjectListConverter(PipPackage)(
            packages,
        ),
        remove_arcana=True,
    )


@attrs.define
class Packages:

    system: list[SystemPackage] = attrs.field(
        factory=list,
        converter=ObjectListConverter(SystemPackage),
        metadata={"serializer": ObjectListConverter.asdict},
    )
    pip: list[PipPackage] = attrs.field(
        factory=list,
        converter=python_package_converter,
        metadata={"serializer": ObjectListConverter.asdict},
    )
    conda: list[CondaPackage] = attrs.field(
        factory=list,
        converter=ObjectListConverter(CondaPackage),
        metadata={"serializer": ObjectListConverter.asdict},
    )
    neurodocker: list[NeurodockerTemplate] = attrs.field(
        factory=list,
        converter=ObjectListConverter(NeurodockerTemplate),
        metadata={"serializer": ObjectListConverter.asdict},
    )


site_pkg_locs = [Path(p).resolve() for p in site.getsitepackages()]



import logging
import sys
import shutil
from pathlib import Path
import json
import typing as ty
from traceback import format_exc
import tempfile
import click
import docker
import docker.errors
from pydra.engine.core import TaskBase
from arcana.core.utils.serialize import (
    package_from_module,
    pydra_asdict,
    ClassResolver,
)
from arcana.core.deploy.image import Metapackage, App
from arcana.core.exceptions import ArcanaBuildError
from arcana.core.data.set.base import Dataset
from arcana.core.data.store import DataStore
from arcana.core.utils.misc import extract_file_from_docker_image, DOCKER_HUB
from .base import cli
from arcana.core.deploy.command import entrypoint_opts

logger = logging.getLogger("arcana")


@cli.group()
def deploy():
    pass


@deploy.command(
    name="make-app",
    help="""Construct and build a dockerfile/apptainer-file for containing a pipeline

SPEC_ROOT is the file system path to the specification to build, or directory
containing multiple specifications

TARGET is the type of image to build, e.g. arcana.xnat.deploy:XnatApp
the target should resolve to a class deriviing from arcana.core.deploy.App.
If it is located under the `arcana.deploy`, then that prefix can be dropped, e.g.
core:App
""",
)
@click.argument("spec_root", type=click.Path(exists=True, path_type=Path))
@click.argument("target", type=str)
@click.option(
    "--registry",
    default=DOCKER_HUB,
    help="The Docker registry to deploy the pipeline to",
)
@click.option(
    "--build-dir",
    default=None,
    type=click.Path(path_type=Path),
    help=(
        "Specify the directory to build the Docker image in. "
        "Defaults to `.build` in the directory containing the "
        "YAML specification"
    ),
)
@click.option(
    "--release",
    default=None,
    nargs=2,
    metavar="<release-name> <release-version>",
    type=str,
    help=("Name of the release for the package as a whole (i.e. for all pipelines)"),
)
@click.option(
    "--tag-latest/--dont-tag-latest",
    default=False,
    type=bool,
    help='whether to tag the release as the "latest" or not',
)
@click.option(
    "--save-manifest",
    default=None,
    type=click.Path(writable=True),
    help="File path at which to save the build manifest",
)
@click.option(
    "--logfile",
    default=None,
    type=click.Path(path_type=Path),
    help="Log output to file instead of stdout",
)
@click.option("--loglevel", default="info", help="The level to display logs at")
@click.option(
    "--use-local-packages/--dont-use-local-packages",
    type=bool,
    default=False,
    help=(
        "Use locally installed Python packages, instead of pulling "
        "them down from PyPI"
    ),
)
@click.option(
    "--install-extras",
    type=str,
    default=None,
    help=(
        "Install extras to use when installing Arcana inside the "
        "container image. Typically only used in tests to provide "
        "'test' extra"
    ),
)
@click.option(
    "--use-test-config/--dont-use-test-config",
    type=bool,  # FIXME: This should be replaced with option to set XNAT CS IP address
    default=False,
    help=(
        "Build the image so that it can be run in Arcana's test "
        "configuration (only for internal use)"
    ),
)
@click.option(
    "--raise-errors/--log-errors",
    type=bool,
    default=False,
    help=("Raise exceptions instead of logging failures"),
)
@click.option(
    "--generate-only/--build",
    type=bool,
    default=False,
    help="Just create the build directory and dockerfile",
)
@click.option(
    "--license",
    type=(str, click.Path(exists=True, path_type=Path)),
    default=(),
    nargs=2,
    metavar="<license-name> <path-to-license-file>",
    multiple=True,
    help=(
        "Licenses provided at build time to be stored in the image (instead of "
        "downloaded at runtime)"
    ),
)
@click.option(
    "--license-to-download",
    type=str,
    default=(),
    multiple=True,
    help=(
        "Specify licenses that are not provided at runtime and instead downloaded "
        "from the data store at runtime in order to satisfy their conditions"
    ),
)
@click.option(
    "--check-registry/--dont-check-registry",
    type=bool,
    default=False,
    help=(
        "Check the registry to see if an existing image with the "
        "same tag is present, and if so whether the specification "
        "matches (and can be skipped) or not (raise an error)"
    ),
)
@click.option(
    "--push/--dont-push",
    type=bool,
    default=False,
    help=("push built images to registry"),
)
@click.option(
    "--clean-up/--dont-clean-up",
    type=bool,
    default=False,
    help=(
        "Remove built images after they are pushed to the registry (requires --push)"
    ),
)
def make_app(
    target,
    spec_root,
    registry,
    release,
    tag_latest,
    save_manifest,
    logfile,
    loglevel,
    build_dir: Path,
    use_local_packages,
    install_extras,
    raise_errors,
    generate_only,
    use_test_config,
    license,
    license_to_download,
    check_registry,
    push,
    clean_up,
):

    if clean_up and not push:
        raise ValueError("'--clean-up' flag requires '--push'")

    if tag_latest and not release:
        raise ValueError("'--tag-latest' flag requires '--release'")

    if isinstance(spec_root, bytes):  # FIXME: This shouldn't be necessary
        spec_root = Path(spec_root.decode("utf-8"))
    if isinstance(build_dir, bytes):  # FIXME: This shouldn't be necessary
        build_dir = Path(build_dir.decode("utf-8"))

    if build_dir is None:
        build_dir = spec_root / ".build"

    if not build_dir.exists():
        build_dir.mkdir()

    install_extras = install_extras.split(",") if install_extras else []

    logging.basicConfig(filename=logfile, level=getattr(logging, loglevel.upper()))

    temp_dir = tempfile.mkdtemp()

    target_cls = ClassResolver(App)(target)

    dc = docker.from_env()

    license_paths = {}
    for lic_name, lic_src in license:
        if isinstance(lic_src, bytes):  # FIXME: This shouldn't be necessary
            lic_src = Path(lic_src.decode("utf-8"))
        license_paths[lic_name] = lic_src

    # Load image specifications from YAML files stored in directory tree

    # Don't error if the modules the task, data stores, data types, etc...
    # aren't present in the build environment
    # FIXME: need to test for this
    with ClassResolver.FALLBACK_TO_STR:
        image_specs = target_cls.load_tree(
            spec_root,
            registry=registry,
            license_paths=license_paths,
            licenses_to_download=set(license_to_download),
        )

    # Check the target registry to see a) if the images with the same tag
    # already exists and b) whether it was built with the same specs
    if check_registry:

        conflicting = {}
        to_build = []
        for image_spec in image_specs:

            extracted_dir = extract_file_from_docker_image(
                image_spec.reference, image_spec.IN_DOCKER_SPEC_PATH
            )
            if extracted_dir is None:
                logger.info(
                    f"Did not find existing image matching {image_spec.reference}"
                )
                changelog = None
            else:
                logger.info(
                    f"Comparing build spec with that of existing image {image_spec.reference}"
                )
                built_spec = image_spec.load(
                    extracted_dir / Path(image_spec.IN_DOCKER_SPEC_PATH).name
                )

                changelog = image_spec.compare_specs(built_spec, check_version=True)

            if changelog is None:
                to_build.append(image_spec)
            elif not changelog:
                logger.info(
                    "Skipping '%s' build as identical image already "
                    "exists in registry"
                )
            else:
                conflicting[image_spec.reference] = changelog

        if conflicting:
            msg = ""
            for tag, changelog in conflicting.items():
                msg += (
                    f"spec for '{tag}' doesn't match the one that was "
                    "used to build the image already in the registry:\n\n"
                    + str(changelog.pretty())
                    + "\n\n\n"
                )

            raise ArcanaBuildError(msg)

        image_specs = to_build

    if release or save_manifest:
        manifest = {
            "package": spec_root.stem,
            "images": [],
        }
        if release:
            manifest["release"] = ":".join(release)

    errors = False

    for image_spec in image_specs:
        spec_build_dir = (
            build_dir / image_spec.loaded_from.relative_to(spec_root)
        ).with_suffix("")
        if spec_build_dir.exists():
            shutil.rmtree(spec_build_dir)
        spec_build_dir.mkdir(parents=True)
        try:
            image_spec.make(
                build_dir=spec_build_dir,
                use_test_config=use_test_config,
                use_local_packages=use_local_packages,
                generate_only=generate_only,
            )
        except Exception:
            if raise_errors:
                raise
            logger.error(
                "Could not build %s pipeline:\n%s", image_spec.reference, format_exc()
            )
            errors = True
            continue
        else:
            click.echo(image_spec.reference)
            logger.info("Successfully built %s pipeline", image_spec.reference)

        if push:
            try:
                dc.api.push(image_spec.reference)
            except Exception:
                if raise_errors:
                    raise
                logger.error(
                    "Could not push '%s':\n\n%s", image_spec.reference, format_exc()
                )
                errors = True
            else:
                logger.info(
                    "Successfully pushed '%s' to registry", image_spec.reference
                )
            if clean_up:
                dc.api.remove_image(image_spec.reference)
                dc.containers.prune()
                dc.images.prune(filters={"dangling": False})
                dc.api.remove_image(image_spec.base_image)
                dc.images.prune(filters={"dangling": False})
                logger.info(
                    "Removed '%s' and pruned dangling images to free up disk space",
                    image_spec.reference,
                )

        if release or save_manifest:
            manifest["images"].append(
                {
                    "name": image_spec.path,
                    "version": image_spec.tag,
                }
            )
    if release:
        metapkg = Metapackage(
            name=release[0],
            version=release[1],
            org=spec_root.stem,
            manifest=manifest,
        )
        metapkg.make(use_local_packages=use_local_packages)
        if push:
            try:
                dc.api.push(metapkg.reference)
            except Exception:
                if raise_errors:
                    raise
                logger.error(
                    "Could not push release metapackage '%s':\n\n%s",
                    metapkg.reference,
                    format_exc(),
                )
                errors = True
            else:
                logger.info(
                    "Successfully pushed release metapackage '%s' to registry",
                    metapkg.reference,
                )

            if tag_latest:
                # Also push release to "latest" tag
                image = dc.images.get(metapkg.reference)
                latest_tag = metapkg.path + ":latest"
                image.reference(latest_tag)

                try:
                    dc.api.push(latest_tag)
                except Exception:
                    if raise_errors:
                        raise
                    logger.error(
                        "Could not push latest tag for release metapackage '%s':\n\n%s",
                        metapkg.path,
                        format_exc(),
                    )
                    errors = True
                else:
                    logger.info(
                        (
                            "Successfully pushed latest tag for release metapackage '%s' "
                            "to registry"
                        ),
                        metapkg.path,
                    )
        if save_manifest:
            with open(save_manifest, "w") as f:
                json.dump(manifest, f, indent="    ")

    shutil.rmtree(temp_dir)
    if errors:
        sys.exit(1)


@deploy.command(
    name="list-images",
    help="""Walk through the specification paths and list tags of the images
that will be build from them.

SPEC_ROOT is the file system path to the specification to build, or directory
containing multiple specifications

DOCKER_ORG is the Docker organisation the images should belong to""",
)
@click.argument("spec_root", type=click.Path(exists=True, path_type=Path))
@click.option(
    "--registry",
    default=None,
    help="The Docker registry to deploy the pipeline to",
)
def list_images(spec_root, registry):

    if isinstance(spec_root, bytes):  # FIXME: This shouldn't be necessary
        spec_root = Path(spec_root.decode("utf-8"))

    for image_spec in App.load_tree(spec_root, registry=registry):
        click.echo(image_spec.reference)


# @deploy.command(
#     name="test",
#     help="""Test container images defined by YAML
# specs

# Arguments
# ---------
# module_path
#     The file system path to the module to build""",
# )
# @click.argument("spec_path", type=click.Path(exists=True, path_type=Path))
# def test(spec_path):
#     # FIXME: Workaround for click 7.x, which improperly handles path_type
#     if type(spec_path) is bytes:
#         spec_path = Path(spec_path.decode("utf-8"))

#     raise NotImplementedError


@deploy.command(
    name="make-docs",
    help="""Build docs for one or more yaml wrappers

SPEC_ROOT is the path of a YAML spec file or directory containing one or more such files.

The generated documentation will be saved to OUTPUT.
""",
)
@click.argument("spec_root", type=click.Path(exists=True, path_type=Path))
@click.argument("output", type=click.Path(path_type=Path))
@click.option(
    "--registry",
    default=DOCKER_HUB,
    help="The Docker registry to deploy the pipeline to",
)
@click.option("--flatten/--no-flatten", default=False)
@click.option("--loglevel", default="warning", help="The level to display logs at")
def make_docs(spec_root, output, registry, flatten, loglevel):
    # FIXME: Workaround for click 7.x, which improperly handles path_type
    if type(spec_root) is bytes:
        spec_root = Path(spec_root.decode("utf-8"))
    if type(output) is bytes:
        output = Path(output.decode("utf-8"))

    logging.basicConfig(level=getattr(logging, loglevel.upper()))

    output.mkdir(parents=True, exist_ok=True)

    with ClassResolver.FALLBACK_TO_STR:
        image_specs = App.load_tree(spec_root, registry=registry)

    for image_spec in image_specs:

        image_spec.autodoc(output, flatten=flatten)
        logging.info("Successfully created docs for %s", image_spec.path)


@deploy.command(
    name="required-packages",
    help="""Detect the Python packages required to run the
specified workflows and return them and their versions""",
)
@click.argument("task_locations", nargs=-1)
def required_packages(task_locations):

    required_modules = set()
    for task_location in task_locations:
        workflow = ClassResolver(TaskBase, alternative_types=[ty.Callable])(
            task_location
        )
        pydra_asdict(workflow, required_modules)

    for pkg in package_from_module(required_modules):
        click.echo(f"{pkg.key}=={pkg.version}")


@deploy.command(
    name="inspect-docker-exec", help="""Extract the executable from a Docker image"""
)
@click.argument("image_tag", type=str)
def inspect_docker_exec(image_tag):
    """Pulls a given Docker image tag and inspects the image to get its
    entrypoint/cmd

    IMAGE_TAG is the tag of the Docker image to inspect"""
    dc = docker.from_env()

    dc.images.pull(image_tag)

    image_attrs = dc.api.inspect_image(image_tag)["Config"]

    executable = image_attrs["Entrypoint"]
    if executable is None:
        executable = image_attrs["Cmd"]

    click.echo(executable)


@deploy.command(
    help="""Displays the changelogs found in the release manifest of a deployment build

MANIFEST_JSON is a JSON file containing a list of container images built in the release
and the commands present in them"""
)
@click.argument("manifest_json", type=click.File())
@click.argument("images", nargs=-1)
def changelog(manifest_json):

    manifest = json.load(manifest_json)

    for entry in manifest["images"]:
        click.echo(
            f"{entry['name']} [{entry['version']}] changes "
            f"from {entry['previous_version']}:\n{entry['changelog']}"
        )


@deploy.command(
    name="install-license",
    help="""Installs a license within a store (i.e. site-wide) or dataset (project-specific)
for use in a deployment pipeline

LICENSE_NAME the name of the license to upload. Must match the name of the license specified
in the deployment specification

SOURCE_FILE path to the license file to upload

INSTALL_LOCATIONS a list of installation locations, which are either the "nickname" of a
store (as saved by `arcana store add`) or the ID of a dataset in form
<store-nickname>//<dataset-id>[@<dataset-name>], where the dataset ID
is either the location of the root directory (for file-system based stores) or the
project ID for managed data repositories.
""",
)
@click.argument("license_name")
@click.argument("source_file", type=click.Path(exists=True, path_type=Path))
@click.argument("install_locations", nargs=-1)
@click.option(
    "--logfile",
    default=None,
    type=click.Path(path_type=Path),
    help="Log output to file instead of stdout",
)
@click.option("--loglevel", default="info", help="The level to display logs at")
def install_license(install_locations, license_name, source_file, logfile, loglevel):

    logging.basicConfig(filename=logfile, level=getattr(logging, loglevel.upper()))

    if isinstance(source_file, bytes):  # FIXME: This shouldn't be necessary
        source_file = Path(source_file.decode("utf-8"))

    if not install_locations:
        install_locations = ["dirtree"]

    for install_loc in install_locations:
        if "//" in install_loc:
            dataset = Dataset.load(install_loc)
            store_name, _, _ = Dataset.parse_id_str(install_loc)
            msg = f"for '{dataset.name}' dataset on {store_name} store"
        else:
            store = DataStore.load(install_loc)
            dataset = store.site_licenses_dataset()
            if dataset is None:
                raise ValueError(
                    f"{install_loc} store doesn't support the installation of site-wide "
                    "licenses, please specify a dataset to install it for"
                )
            msg = f"site-wide on {install_loc} store"

        dataset.install_license(license_name, source_file)
        logger.info("Successfully installed '%s' license %s", license_name, msg)


@deploy.command(
    name="pipeline-entrypoint",
    help="""Loads/creates a dataset, then applies and launches a pipeline
in a single command. To be used within the command configuration of an XNAT
Container Service ready Docker image.

DATASET_LOCATOR string containing the nickname of the data store, the ID of the
dataset (e.g. XNAT project ID or file-system directory) and the dataset's name
in the format <store-nickname>//<dataset-id>[@<dataset-name>]

""",
)
@click.argument("dataset_locator")
@entrypoint_opts.data_columns
@entrypoint_opts.parameterisation
@entrypoint_opts.execution
@entrypoint_opts.dataset_config
@entrypoint_opts.debugging
def pipeline_entrypoint(
    dataset_locator,
    spec_path,
    **kwargs,
):

    image_spec = App.load(spec_path)

    image_spec.command.execute(
        dataset_locator,
        **kwargs,
    )



from arcana.core.data import Clinical


def test_is_parent():
    assert not Clinical.session.is_parent(Clinical.session)
    assert not Clinical.session.is_parent(Clinical.subject)
    assert not Clinical.session.is_parent(Clinical.member)
    assert not Clinical.session.is_parent(Clinical.group)
    assert not Clinical.session.is_parent(Clinical.timepoint)
    assert not Clinical.session.is_parent(Clinical.batch)
    assert not Clinical.session.is_parent(Clinical.matchedpoint)
    assert not Clinical.session.is_parent(Clinical.dataset)

    assert Clinical.member.is_parent(Clinical.session)
    assert Clinical.member.is_parent(Clinical.subject)
    assert not Clinical.member.is_parent(Clinical.member)
    assert not Clinical.member.is_parent(Clinical.group)
    assert not Clinical.member.is_parent(Clinical.timepoint)
    assert not Clinical.member.is_parent(Clinical.batch)
    assert Clinical.member.is_parent(Clinical.matchedpoint)
    assert not Clinical.member.is_parent(Clinical.dataset)

    assert Clinical.group.is_parent(Clinical.session)
    assert Clinical.group.is_parent(Clinical.subject)
    assert not Clinical.group.is_parent(Clinical.member)
    assert not Clinical.group.is_parent(Clinical.group)
    assert not Clinical.group.is_parent(Clinical.timepoint)
    assert Clinical.group.is_parent(Clinical.batch)
    assert not Clinical.group.is_parent(Clinical.matchedpoint)
    assert not Clinical.group.is_parent(Clinical.dataset)

    assert Clinical.timepoint.is_parent(Clinical.session)
    assert not Clinical.timepoint.is_parent(Clinical.subject)
    assert not Clinical.timepoint.is_parent(Clinical.member)
    assert not Clinical.timepoint.is_parent(Clinical.group)
    assert not Clinical.timepoint.is_parent(Clinical.timepoint)
    assert Clinical.timepoint.is_parent(Clinical.batch)
    assert Clinical.timepoint.is_parent(Clinical.matchedpoint)
    assert not Clinical.timepoint.is_parent(Clinical.dataset)

    assert Clinical.subject.is_parent(Clinical.session)
    assert not Clinical.subject.is_parent(Clinical.subject)
    assert not Clinical.subject.is_parent(Clinical.member)
    assert not Clinical.subject.is_parent(Clinical.group)
    assert not Clinical.subject.is_parent(Clinical.timepoint)
    assert not Clinical.subject.is_parent(Clinical.batch)
    assert not Clinical.subject.is_parent(Clinical.matchedpoint)
    assert not Clinical.subject.is_parent(Clinical.dataset)

    assert Clinical.batch.is_parent(Clinical.session)
    assert not Clinical.batch.is_parent(Clinical.subject)
    assert not Clinical.batch.is_parent(Clinical.member)
    assert not Clinical.batch.is_parent(Clinical.group)
    assert not Clinical.batch.is_parent(Clinical.timepoint)
    assert not Clinical.batch.is_parent(Clinical.batch)
    assert not Clinical.batch.is_parent(Clinical.matchedpoint)
    assert not Clinical.batch.is_parent(Clinical.dataset)

    assert Clinical.matchedpoint.is_parent(Clinical.session)
    assert not Clinical.matchedpoint.is_parent(Clinical.subject)
    assert not Clinical.matchedpoint.is_parent(Clinical.member)
    assert not Clinical.matchedpoint.is_parent(Clinical.group)
    assert not Clinical.matchedpoint.is_parent(Clinical.timepoint)
    assert not Clinical.matchedpoint.is_parent(Clinical.batch)
    assert not Clinical.matchedpoint.is_parent(Clinical.matchedpoint)
    assert not Clinical.matchedpoint.is_parent(Clinical.dataset)

    assert Clinical.dataset.is_parent(Clinical.session)
    assert Clinical.dataset.is_parent(Clinical.subject)
    assert Clinical.dataset.is_parent(Clinical.member)
    assert Clinical.dataset.is_parent(Clinical.group)
    assert Clinical.dataset.is_parent(Clinical.timepoint)
    assert Clinical.dataset.is_parent(Clinical.batch)
    assert Clinical.dataset.is_parent(Clinical.matchedpoint)
    assert not Clinical.dataset.is_parent(Clinical.dataset)



from __future__ import annotations
import typing as ty
import itertools
from copy import copy
from collections import defaultdict
import operator as operator_module
import attrs
from attrs.converters import default_if_none
from ..data.space import DataSpace
from .salience import CheckSalience, ColumnSalience, ParameterSalience
from ..utils.misc import ARCANA_SPEC
from arcana.core.exceptions import ArcanaDesignError


@attrs.define
class BaseAttr:

    name: str = None  # Default to None, as will be set later after class is initialised
    type: type = None
    desc: str = None
    inherited: bool = False
    defined_in: ty.Tuple[type] = ()
    modified: ty.Tuple[ty.Tuple[str, ty.Any]] = ()
    mapped_from: ty.Tuple[
        str, str or ty.Tuple
    ] or None = None  # sub-analysis name, column name
    metadata: dict = attrs.field(
        factory=dict, converter=default_if_none(default=attrs.Factory(dict))
    )


@attrs.define(frozen=True)
class ColumnSpec(BaseAttr):
    """Specifies a column that the analysis can add when it is applied to a dataset"""

    row_frequency: DataSpace = None
    salience: ColumnSalience = ColumnSalience.default()

    def select_pipeline_builders(self, analysis, dataset):
        candidates = [
            p for p in analysis.__spec__.pipeline_builders if self.name in p.outputs
        ]
        selected = [
            m
            for m in candidates
            if (
                m.condition is not None
                and m.condition.evaluate(self, analysis, dataset)
            )
        ]
        # Check for defaults
        if not selected:
            selected = [p for p in candidates if p.condition is None]
        # Select pipeline builders from subanalysis if present
        if not selected and self.mapped_from is not None:
            subanalysis = getattr(analysis, self.mapped_from[0])
            sub_column_spec = subanalysis.__spec__.column(self.mapped_from[1])
            selected = sub_column_spec.select_pipeline_builders(subanalysis, dataset)

        if not selected:
            raise ArcanaDesignError(
                "Could not find any potential pipeline builders with conditions that "
                "match the current analysis parameterisation and provided dataset. "
                f"All candidates are: {candidates}"
            )
        # Check to see whether there are pipelines with the same switch
        all_switches = [p.switch for p in selected]
        if with_duplicate_switches := [
            p for p in selected if all_switches.count(p.switch)
        ]:
            raise ArcanaDesignError(
                "Multiple potential pipelines match criteria for the given analysis "
                f"configuration and provided dataset: {with_duplicate_switches}"
            )
        return selected

    def to_attrs_field(self):
        return attrs.field(
            default=None,
            metadata={ARCANA_SPEC: self},
        )


@attrs.define(frozen=True)
class Parameter(BaseAttr):
    """Specifies a free parameter of an analysis"""

    default: int or float or str or ty.Tuple[int] or ty.Tuple[float] or ty.Tuple[
        str
    ] = attrs.field(default=None)
    salience: ParameterSalience = None
    choices: ty.Union[
        ty.Tuple[int], ty.Tuple[float], ty.Tuple[str], None
    ] = attrs.field(default=None)
    lower_bound: ty.Union[int, float, None] = None
    upper_bound: ty.Union[int, float, None] = None

    @default.validator
    def default_validator(self, _, default):
        if default is None:
            if self.salience != ParameterSalience.required:
                raise ValueError(
                    f"Default value for '{self.name}' parameter must be provided unless "
                    f"parameter salience is set to'{ParameterSalience.required}'"
                )
        else:
            if self.lower_bound is not None and default < self.lower_bound:
                raise ValueError(
                    f"Default provided to '{self.name}' ({default}) is lower than "
                    f"lower bound ({self.lower_bound})"
                )
            if self.upper_bound is not None and default > self.upper_bound:
                raise ValueError(
                    f"Default provided to '{self.name}' ({default}) is higher than "
                    f"upper bound ({self.upper_bound})"
                )

    @choices.validator
    def choices_validator(self, _, choices):
        if choices is not None:
            if self.upper_bound is not None or self.lower_bound is not None:
                raise ArcanaDesignError(
                    f"Cannot specify lower ({self.lower_bound}) or upper "
                    f"({self.upper_bound}) bound in conjunction with 'choices' arg "
                    f"({choices})"
                )

    def to_attrs_field(self):
        return attrs.field(
            default=self.default,
            validator=_parameter_validator,
            metadata={ARCANA_SPEC: self},
        )


@attrs.define(frozen=True)
class SubanalysisSpec(BaseAttr):
    """Specifies a "sub-analysis" component, when composing an analysis of several
    predefined analyses"""

    mappings: ty.Tuple[
        str, str
    ] = ()  # to name in subanalysis, from name in analysis class

    def mapping(self, name):
        try:
            mapped_name = next(m[1] for m in self.mappings if name == m[0])
        except StopIteration:
            raise KeyError(f"No mapping from '{name}' in sub-analysis: {self.mappings}")
        return mapped_name

    def to_attrs_field(self):
        return attrs.field(
            metadata={ARCANA_SPEC: self},
            factory=dict,
        )


@attrs.define(frozen=True)
class Operation:
    """Defines logical expressions used in specifying conditions when different versions
    of pipelines will run"""

    operator: str
    operands: ty.Tuple[str]

    def evaluate(self, analysis, dataset):
        operands = [o.evaluate(analysis, dataset) for o in self.operands]
        if self.operator == "value_of":
            assert len(operands) == 1
            val = getattr(analysis, operands[0])
        elif self.operator == "is_provided":
            assert len(operands) <= 2
            column_name = getattr(analysis, operands[0])
            if column_name is None:
                val = False
            else:
                column = dataset[column_name]
                if len(operands) == 2:
                    in_format = operands[1]
                    val = column.datatype is in_format or issubclass(
                        column.datatype, in_format
                    )
                else:
                    val = True
        else:
            val = getattr(operator_module, self.operator)(*operands)
        return val


@attrs.define(frozen=True)
class BaseMethod:

    name: str
    desc: str
    inputs: ty.Tuple[str]
    parameters: ty.Tuple[str]
    method: ty.Callable
    defined_in: ty.Tuple[type]


@attrs.define(frozen=True)
class Switch(BaseMethod):
    """Specifies a "switch" point at which the processing can bifurcate to handle two
    separate types of input streams"""

    pass


@attrs.define(frozen=True)
class PipelineConstructor(BaseMethod):
    """Specifies a method that is used to add nodes in the construction of a pipeline
    that is able to generate data for sink columns under certain conditions"""

    outputs: ty.Tuple[str]
    condition: ty.Union[Operation, None] = None
    switch: ty.Union[Switch, None] = None


@attrs.define(frozen=True)
class Check(BaseMethod):
    """Specifies a quality-control check that can be run on generated derivatives to
    assess the probability that they have failed"""

    column: ColumnSpec
    salience: CheckSalience = CheckSalience.default()


def unique_names(inst, attr, val):
    names = [v.name for v in val]
    if duplicates := [v for v in val if names.count(v.name) > 1]:
        raise ValueError(f"Duplicate names found in provided tuple: {duplicates}")


@attrs.define(frozen=True)
class AnalysisSpec:
    """Specifies all the components of the analysis class"""

    space: type
    column_specs: ty.Tuple[ColumnSpec] = attrs.field(validator=unique_names)
    pipeline_builders: ty.Tuple[PipelineConstructor] = attrs.field(
        validator=unique_names
    )
    parameters: ty.Tuple[Parameter] = attrs.field(validator=unique_names)
    switches: ty.Tuple[Switch] = attrs.field(validator=unique_names)
    checks: ty.Tuple[Check] = attrs.field(validator=unique_names)
    subanalysis_specs: ty.Tuple[SubanalysisSpec] = attrs.field(validator=unique_names)

    @property
    def column_names(self):
        return (c.name for c in self.column_specs)

    @property
    def parameter_names(self):
        return (p.name for p in self.parameters)

    @property
    def pipeline_names(self):
        return (p.name for p in self.pipeline_builders)

    @property
    def switch_names(self):
        return (s.name for s in self.switches)

    @property
    def check_names(self):
        return (c.name for c in self.checks)

    @property
    def subanalysis_names(self):
        return (s.name for s in self.subanalysis_specs)

    def column_spec(self, name):
        try:
            return next(c for c in self.column_specs if c.name == name)
        except StopIteration:
            raise KeyError(f"No column spec named '{name}' in {self}")

    def parameter(self, name):
        try:
            return next(p for p in self.parameters if p.name == name)
        except StopIteration:
            raise KeyError(f"No parameter named '{name}' in {self}")

    def subanalysis_spec(self, name):
        try:
            return next(s for s in self.subanalysis_specs if s.name == name)
        except StopIteration:
            raise KeyError(f"No subanalysis spec named '{name}' in {self}")

    def pipeline_builder(self, name):
        try:
            return next(p for p in self.pipeline_builders if p.name == name)
        except StopIteration:
            raise KeyError(f"No pipeline builder named '{name}' in {self}")

    def switch(self, name):
        try:
            return next(s for s in self.switches if s.name == name)
        except StopIteration:
            raise KeyError(f"No switches named '{name}' in {self}")

    def check(self, name):
        try:
            return next(c for c in self.checks if c.name == name)
        except StopIteration:
            raise KeyError(f"No checks named '{name}' in {self}")

    def member(self, name):
        try:
            return next(m for m in self.members() if m.name == name)
        except StopIteration:
            raise KeyError(f"No member named '{name}' in {self}")

    def members(self):
        return itertools.chain(
            self.column_specs, self.parameters, self.subanalysis_specs
        )

    def column_checks(self, column_name):
        "Return all checks for a given column"
        return (c for c in self.checks if c.column == column_name)

    @column_specs.validator
    def column_specs_validator(self, _, column_specs):
        for column_spec in column_specs:
            sorted_by_cond = defaultdict(list)
            for pipe_spec in self.pipeline_builders:
                if column_spec.name in pipe_spec.outputs:
                    sorted_by_cond[(pipe_spec.condition, pipe_spec.switch)].append(
                        pipe_spec
                    )
            if duplicated := [
                (c, d) for (c, d) in sorted_by_cond.items() if len(d) > 1
            ]:
                raise ArcanaDesignError(
                    f"Multiple pipelines provide outputs for '{column_spec.name}' under "
                    "matching conditions - \n"
                    + "\n".join(
                        f"Condition: {cond[0]}, Switch: {cond[1]} - "
                        + ", ".join(str(p) for p in dups)
                        for cond, dups in duplicated
                    )
                )
            if not sorted_by_cond and not column_spec.mapped_from:
                inputs_to = [
                    p for p in self.pipeline_builders if column_spec.name in p.inputs
                ]
                if not inputs_to:
                    raise ArcanaDesignError(
                        f"'{column_spec.name}' is neither an input nor output to any pipeline"
                    )
                if column_spec.salience.level <= ColumnSalience.publication.level:
                    raise ArcanaDesignError(
                        f"'{column_spec.name}' is not generated by any pipeline yet its salience "
                        f"is not specified as 'raw' or 'primary'"
                    )

    @pipeline_builders.validator
    def pipeline_builders_validator(self, _, pipeline_builders):
        for pipeline_builder in pipeline_builders:
            if missing_outputs := [
                o for o in pipeline_builder.outputs if o not in self.column_names
            ]:
                raise ArcanaDesignError(
                    f"'{pipeline_builder.name}' pipeline outputs to unknown columns: {missing_outputs}"
                )


@attrs.define
class _Inherited:

    to_overwrite: ty.Dict[str, ty.Any]
    resolved_to: str = None
    name: str = None

    def resolve(self, name, klass):
        """Resolve to columns and parameters in the specified class

        Parameters
        ----------
        name : str
            the name of the attribute in the class to resolve the inherited attribute to
        klass : type
            the initial class to be transformed into an analysis class
        """
        self.name = name
        defining_class = None
        for base in klass.__mro__[1:-1]:  # skip current class and base "object" class
            if name in base.__dict__:
                defining_class = base
        if not defining_class:
            raise AttributeError(
                f"Supers of {klass} have no attribute named '{name}' to inherit"
            )
        attr_to_inherit = getattr(attrs.fields(defining_class), name).metadata[
            ARCANA_SPEC
        ]
        kwargs = copy(self.to_overwrite)
        if self.to_overwrite:
            kwargs["modified"] = attr_to_inherit.modified + (
                tuple(self.to_overwrite.items()),
            )
        kwargs["inherited"] = True
        resolved = attrs.evolve(attr_to_inherit, **kwargs)
        # Return the resolved attribute and its type annotation
        return resolved, resolved.type


@attrs.define
class _MappedFrom:

    subanalysis_name: str
    attr_name: str
    to_overwrite: ty.Dict[str, ty.Any]
    resolved_to: str = None
    name: str = None

    def resolve(self, name, klass):
        """Resolve to a column temporary attribute to be transformed into a attribute
        of the analysis class

        Parameters
        ----------
        name : str
            the name of the attribute in the class to resolve the inherited attribute to
        klass : type
            the initial class to be transformed into an analysis class
        """
        self.name = name
        analysis_class = klass.__annotations__[self.subanalysis_name]
        analysis_spec = analysis_class.__spec__
        # Get the Attribute in the subanalysis class
        try:
            attr_spec = analysis_spec.member(self.attr_name)
        except KeyError:
            raise ArcanaDesignError(
                f"No member attribute named '{self.attr_name}' in subanalysis "
                f"'{self.subanalysis_name}' ({analysis_class}): "
                + str([a.name for a in analysis_class.__attrs_attrs__])
            )
        kwargs = copy(self.to_overwrite)
        kwargs["mapped_from"] = (self.subanalysis_name, self.attr_name)
        kwargs["modified"] = attr_spec.modified + (tuple(self.to_overwrite.items()),)
        resolved = attrs.evolve(attr_spec, **kwargs)
        return resolved, attr_spec.type


@attrs.define
class _UnresolvedOp:
    """An operation within a conditional expression that hasn't been resolved"""

    operator: str
    operands: ty.Tuple[str]

    def resolve(self, klass, column_specs, parameters):
        """Resolves counting attribute operands to the names of attributes in the class

        Parameters
        ----------
        klass : type
            the class being wrapped by attrs.define
        column_specs : list[ColumnSpec]
            the column specs defined in the class
        parameters : list[Parameter]
            the parameters defined in the class

        Return
        ------
        _Op
            the operator with counting-attributes by attrs resolved to attribute names
        """
        resolved = []
        for operand in self.operands:
            if isinstance(operand, _UnresolvedOp):
                operand = operand.resolve(klass, column_specs, parameters)
            else:
                try:
                    operand = operand.resolved_to
                except AttributeError:
                    pass
            resolved.append(operand)
        if self.operator == "value_of":
            assert len(resolved) == 1
            if resolved[0] not in parameters:
                raise ValueError(
                    f"'value_of' can only be used on parameter attributes not '{operand}'"
                )
        elif self.operator == "is_provided":
            assert len(resolved) <= 2
            if resolved[0] not in column_specs:
                raise ValueError(
                    f"'is_provided' can only be used on column specs not '{operand}'"
                )
        return Operation(
            self.operator, tuple(a.name if hasattr(a, "name") else a for a in resolved)
        )

    def __eq__(self, o):
        return _UnresolvedOp("eq", (self, o))

    def __ne__(self, o):
        return _UnresolvedOp("ne", (self, o))

    def __lt__(self, o):
        return _UnresolvedOp("lt", (self, o))

    def __le__(self, o):
        return _UnresolvedOp("le", (self, o))

    def __gt__(self, o):
        return _UnresolvedOp("gt", (self, o))

    def __ge__(self, o):
        return _UnresolvedOp("ge", (self, o))

    def __and__(self, o):
        return _UnresolvedOp("and_", (self, o))

    def __or__(self, o):
        return _UnresolvedOp("or_", (self, o))

    def __invert__(self):
        return _UnresolvedOp("invert_", (self,))


def _parameter_validator(self, attr, val):
    spec = attr.metadata[ARCANA_SPEC]
    if spec.salience is ParameterSalience.required and val is None:
        raise ValueError(
            f"A value needs to be provided to required parameter '{attr.name}' in {self}"
        )
    if spec.choices is not None and val not in spec.choices:
        val_str = f"'{val}'" if isinstance(val, str) else val
        raise ValueError(
            f"{val_str} is not a valid value for '{attr.name}' parameter in {self}, "
            f"valid choices are {spec.choices}"
        )
    if not (
        (spec.lower_bound is None or val >= spec.lower_bound)
        and (spec.upper_bound is None or val <= spec.upper_bound)
    ):
        raise ValueError(
            f"Value of '{attr.name}' ({val}) is not within the specified bounds: "
            f"{spec.lower_bound} - {spec.upper_bound} in {self}"
        )


# def _column_validator(self, attr, val):
#     if val not in self._dataset.columns():
#         raise ValueError(
#             f"Unrecognised column name '{val}' provided as mapping for {attr.name} "
#             "column spec")



from __future__ import annotations
from typing import Union, Dict, Tuple
import shutil
import re
import yaml
import tempfile
from pathlib import Path
import attrs
import pytest
import docker
from arcana.core.cli.deploy import (
    make_app,
    make_docs,
)
from arcana.core.utils.misc import show_cli_trace
from arcana.core.exceptions import ArcanaBuildError


def test_deploy_make_app_cli(command_spec, cli_runner, work_dir):

    DOCKER_ORG = "testorg"
    DOCKER_REGISTRY = "test.registry.org"
    IMAGE_GROUP_NAME = "testpkg"

    concatenate_spec = {
        "command": command_spec,
        "version": "1.0",
        "build_iteration": "1",
        "packages": {
            "system": ["vim"],  # just to test it out
            "pip": {"pydra": None},  # just to test it out
        },
        "authors": [{"name": "Some One", "email": "some.one@an.email.org"}],
        "info_url": "http://concatenate.readthefakedocs.io",
        "description": "a test image spec",
    }

    build_dir = work_dir / "build"
    build_dir.mkdir()
    spec_path = work_dir / DOCKER_ORG
    sub_dir = spec_path / IMAGE_GROUP_NAME
    sub_dir.mkdir(parents=True)
    with open(sub_dir / "concatenate.yml", "w") as f:
        yaml.dump(concatenate_spec, f)

    result = cli_runner(
        make_app,
        [
            str(spec_path),
            "core:App",
            "--build-dir",
            str(build_dir),
            "--registry",
            DOCKER_REGISTRY,
            "--loglevel",
            "warning",
            "--use-local-packages",
            "--install-extras",
            "test",
            "--raise-errors",
            "--use-test-config",
            "--dont-check-registry",
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    tag = result.output.strip().splitlines()[-1]
    assert tag == f"{DOCKER_REGISTRY}/{DOCKER_ORG}/{IMAGE_GROUP_NAME}.concatenate:1.0-1"

    # Clean up the built image
    dc = docker.from_env()
    dc.images.remove(tag)


def test_deploy_remake_app_cli(command_spec, docker_registry, cli_runner, run_prefix):
    """Tests the check to see whether"""

    IMAGE_GROUP_NAME = "testpkg-rebuild" + run_prefix

    def build_spec(spec, **kwargs):
        work_dir = Path(tempfile.mkdtemp())
        build_dir = work_dir / "build"
        build_dir.mkdir()
        spec_path = work_dir / "testorg"
        sub_dir = spec_path / IMAGE_GROUP_NAME
        sub_dir.mkdir(parents=True)
        with open(sub_dir / "concatenate.yml", "w") as f:
            yaml.dump(spec, f)

        result = cli_runner(
            make_app,
            [
                str(spec_path),
                "core:App",
                "--build-dir",
                str(build_dir),
                "--registry",
                docker_registry,
                "--loglevel",
                "warning",
                "--use-local-packages",
                "--install-extras",
                "test",
                "--raise-errors",
                "--check-registry",
                "--use-test-config",
            ],
            **kwargs,
        )
        return result

    concatenate_spec = {
        "command": command_spec,
        "version": "1.0",
        "build_iteration": "1",
        "packages": {"system": ["vim"]},
        "description": "a test image",
        "name": "test_deploy_rebuild_cli",
        "authors": [{"name": "Some One", "email": "some.one@an.email.org"}],
        "info_url": "http://concatenate.readthefakedocs.io",
    }

    # Build a basic image
    result = build_spec(concatenate_spec)
    assert result.exit_code == 0, show_cli_trace(result)
    assert result.output
    tag = result.output.strip().splitlines()[-1]
    try:
        dc = docker.from_env()
        dc.api.push(tag)

        # FIXME: Need to ensure that logs are captured properly then we can test this
        # result = build_spec(concatenate_spec)
        # assert "Skipping" in result.output

        # Modify the spec so it doesn't match the original that has just been
        # built (but don't increment the version number -> image tag so there
        # is a clash)
        concatenate_spec["packages"] = {"system": ["vim", "git"]}

        with pytest.raises(ArcanaBuildError) as excinfo:
            build_spec(concatenate_spec, catch_exceptions=False)

        assert "doesn't match the one that was used to build the image" in str(
            excinfo.value
        )

        # Increment the version number to avoid the clash
        concatenate_spec["build_iteration"] = "2"

        result = build_spec(concatenate_spec)
        assert result.exit_code == 0, show_cli_trace(result)
        rebuilt_tag = result.output.strip().splitlines()[-1]
        dc.images.remove(rebuilt_tag)
    finally:
        # Clean up the built images
        dc.images.remove(tag)


@attrs.define
class DocsFixture:

    yaml_src: str
    markdown: str
    licenses_to_provide: list[str] = attrs.field(factory=list)


docs_fixtures = {
    "simple": DocsFixture(
        """
version: &version '0.16.1'
authors:
  - name: author_name
    email: author@email.org
base_image:
  name: abc
  tag: *version
  package_manager: apt
info_url: https://example.com
description: >-
  a test of the YAML join functionality
command:
  task: arcana.testing.tasks:identity_file
  row_frequency: common:Samples[sample]
  inputs:
    in_file:
      datatype: fileformats.text:Plain
      help_string: the input file
  outputs:
    out_file:
      datatype: fileformats.text:Plain
      help_string: the output file
    """.strip(),
        """
---
source_file: spec.yaml
title: spec
weight: 10

---

## Package Info
|Key|Value|
|---|-----|
|Name|spec|
|App version|0.16.1|
|Build iteration|0|
|Base image|`abc:0.16.1`|
|Maintainer|author_name (author@email.org)|
|Info URL|https://example.com|
|Short description|a test of the YAML join functionality|

a test of the YAML join functionality

## Command
|Key|Value|
|---|-----|
|Task|arcana.testing.tasks:identity_file|
|Operates on|sample|
#### Inputs
|Name|Required data-type|Default column data-type|Description|
|----|------------------|------------------------|-----------|
|`in_file`|<span data-toggle="tooltip" data-placement="bottom" title="text/plain" aria-label="text/plain">text/plain</span>|<span data-toggle="tooltip" data-placement="bottom" title="text/plain" aria-label="text/plain">text/plain</span>|the input file|

#### Outputs
|Name|Required data-type|Default column data-type|Description|
|----|------------------|------------------------|-----------|
|`out_file`|<span data-toggle="tooltip" data-placement="bottom" title="text/plain" aria-label="text/plain">text/plain</span>|<span data-toggle="tooltip" data-placement="bottom" title="text/plain" aria-label="text/plain">text/plain</span>|the output file|

#### Parameters
|Name|Data type|Description|
|----|---------|-----------|
""".strip(),
    ),
    "full": DocsFixture(
        """
version: &version '0.16.1'
build_iteration: '10'
authors:
  - name: author_name
    email: author@email.org
base_image:
  name: abc
  tag: *version
  package_manager: yum
description: a description
long_description: >-
  a longer description
known_issues:
  - url: https://example.com
info_url: https://example.com
packages:
  system:
    vim: 99.1
    git:
  pip:
    - pydra
    - pydra-dcm2niix
  neurodocker:
    dcm2niix: v1.0.20201102
licenses:
  freesurfer:
    destination: /opt/freesurfer/license.txt
    description: >
      license description
    info_url: http://path.to.license.provider.org/licenses
command:
    task: bids:bids_app
    inputs:
      T1w:
        configuration:
          path: anat/T1w
        datatype: fileformats.medimage:NiftiGzX
        help_string: "T1-weighted anatomical scan"
        default_column:
          datatype: fileformats.medimage:DicomSet
      T2w:
        configuration:
          path: anat/T2w
        datatype: fileformats.medimage:NiftiGzX
        help_string: "T2-weighted anatomical scan"
        default_column:
          datatype: fileformats.medimage:DicomSet
      fMRI:
        datatype: fileformats.medimage:NiftiGzX
        help_string: "functional MRI"
        configuration:
          path: func/bold/task=rest
        default_column:
          datatype: fileformats.medimage:DicomSet
    outputs:
      mriqc:
        datatype: fileformats.generic:Directory
        help_string: "MRIQC output directory"
        configuration:
          path: mriqc
    parameters:
      fmriprep_flags:
        field: flags
        datatype: str
        help_string: description of flags param
    row_frequency: core:Clinical[session]
    configuration:
      executable: /usr/local/miniconda/bin/mriqc
      dataset: /work/bids-dataset
      app_output_dir: /work/bids-app-output
    """.strip(),
        """
---
source_file: /var/folders/mz/yn83q2fd3s758w1j75d2nnw80000gn/T/tmp47_dxmyq/specs/spec.yaml
title: spec
weight: 10

---

## Package Info
|Key|Value|
|---|-----|
|Name|spec|
|App version|0.16.1|
|Build iteration|10|
|Base image|`abc:0.16.1`|
|Maintainer|author_name (author@email.org)|
|Info URL|https://example.com|
|Short description|a description|
|Known issues|https://example.com|

a longer description

### Required licenses
|Name|URL|Description|
|----|---|-----------|
|freesurfer|`http://path.to.license.provider.org/licenses`|license description|

## Command
|Key|Value|
|---|-----|
|Task|bids:bids_app|
|Operates on|session|
#### Inputs
|Name|Required data-type|Default column data-type|Description|
|----|------------------|------------------------|-----------|
|`T1w`|<span data-toggle="tooltip" data-placement="bottom" title="medimage/nifti-gz-x" aria-label="medimage/nifti-gz-x">medimage/nifti-gz-x</span>|<span data-toggle="tooltip" data-placement="bottom" title="medimage/dicom-set" aria-label="medimage/dicom-set">medimage/dicom-set</span>|T1-weighted anatomical scan|
|`T2w`|<span data-toggle="tooltip" data-placement="bottom" title="medimage/nifti-gz-x" aria-label="medimage/nifti-gz-x">medimage/nifti-gz-x</span>|<span data-toggle="tooltip" data-placement="bottom" title="medimage/dicom-set" aria-label="medimage/dicom-set">medimage/dicom-set</span>|T2-weighted anatomical scan|
|`fMRI`|<span data-toggle="tooltip" data-placement="bottom" title="medimage/nifti-gz-x" aria-label="medimage/nifti-gz-x">medimage/nifti-gz-x</span>|<span data-toggle="tooltip" data-placement="bottom" title="medimage/dicom-set" aria-label="medimage/dicom-set">medimage/dicom-set</span>|functional MRI|

#### Outputs
|Name|Required data-type|Default column data-type|Description|
|----|------------------|------------------------|-----------|
|`mriqc`|<span data-toggle="tooltip" data-placement="bottom" title="generic/directory" aria-label="generic/directory">generic/directory</span>|<span data-toggle="tooltip" data-placement="bottom" title="generic/directory" aria-label="generic/directory">generic/directory</span>|MRIQC output directory|

#### Parameters
|Name|Data type|Description|
|----|---------|-----------|
|`fmriprep_flags`|`str`|description of flags param|
""".strip(),
        ["freesurfer"],
    ),
}


@pytest.mark.parametrize("fixture", docs_fixtures.items(), ids=lambda x: x[0])
def test_make_docs_cli(
    cli_runner, run_prefix, work_dir: Path, fixture: Tuple[str, DocsFixture]
):
    fixture_name, fixture_content = fixture

    # TODO handle multiple 'files' in a fixture
    print(f"Processing fixture: {fixture_name!r}")
    output = _make_docs(cli_runner, work_dir, fixture_content.yaml_src)

    strip_source_file_re = re.compile(r"source_file:.*")

    stripped_output = strip_source_file_re.sub("", output)
    stripped_reference = strip_source_file_re.sub("", fixture_content.markdown)

    assert (
        stripped_output == stripped_reference
    ), f"Fixture {fixture_name!r} didn't match output"


def _make_docs(
    cli_runner,
    work_dir: Path,
    docs: Union[str, Dict[str, str]],
    *args,
    flatten: bool = None,
) -> Union[str, Dict[str, str]]:
    out_dir = work_dir / "out"
    specs_dir = work_dir / "specs"
    if specs_dir.exists():
        shutil.rmtree(specs_dir)
    specs_dir.mkdir()

    if type(docs) is str:
        (specs_dir / "spec.yaml").write_text(docs)
    else:
        for name, content in docs.items():
            path = specs_dir / name
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_text(content)

    result = cli_runner(
        make_docs,
        [
            specs_dir.as_posix(),
            out_dir.as_posix(),
        ]
        + (["--flatten" if flatten else "--no-flatten"] if flatten is not None else [])
        + list(args),
    )

    assert result.exit_code == 0, show_cli_trace(result)

    if type(docs) is str:
        return (out_dir / "spec.md").read_text().strip()
    else:
        return {
            file.relative_to(out_dir).as_posix(): file.read_text().strip()
            for file in out_dir.glob("*.md")
        }



from __future__ import annotations
import logging
import typing as ty
import re
import attrs
import attrs.filters
from arcana.core.utils.misc import NestedContext
from arcana.core.data.space import DataSpace
from arcana.core.exceptions import (
    ArcanaNameError,
    ArcanaDataTreeConstructionError,
)
from .row import DataRow

if ty.TYPE_CHECKING:  # pragma: no cover
    from .set.base import Dataset


logger = logging.getLogger("arcana")


@attrs.define
class DataTree(NestedContext):

    dataset: Dataset = None
    root: DataRow = None

    def enter(self):
        assert self.root is None
        self._set_root()
        self.dataset.store.populate_tree(self)

    def exit(self):
        self.root = None

    @property
    def dataset_id(self):
        return self.dataset.id

    @property
    def hierarchy(self):
        return self.dataset.hierarchy

    def add_leaf(
        self, tree_path, metadata: dict[str, dict[str, str]] = None
    ) -> tuple[DataRow, list[str]]:
        """Creates a new row at a the path down the tree of the dataset as
        well as all "parent" rows upstream in the data tree

        Parameters
        ----------
        tree_path : list[str]
            The sequence of labels for each layer in the hierarchy of the
            dataset leading to the current row.
        metadata : dict[str, dict[str, str]]
            metadata passed to ``DataStore.infer_ids()`` used to infer IDs not directly
            represented in the hierarchy of the data tree.

        Returns
        -------
        row : DataRow or None
            the added row if it is not excluded, None if it was excluded
        exclusions : list[str]
            the list of frequencies that caused the leaf to be excluded (empty if it
            was added) according to the the exclusion criteria provided to the dataset
            (see ``Dataset.include`` and ``Dataset.exclude``)

        Raises
        ------
        ArcanaBadlyFormattedIDError
            raised if one of the IDs doesn't match the pattern in the
            `id_patterns`
        ArcanaDataTreeConstructionError
            raised if one of the groups specified in the ID inference reg-ex
            doesn't match a valid row_frequency in the data dimensions
        """
        if self.root is None:
            self._set_root()
        if metadata is None:
            metadata = {}
        # Get basis frequencies covered at the given depth of the
        if len(tree_path) != len(self.dataset.hierarchy):
            raise ArcanaDataTreeConstructionError(
                f"Tree path ({tree_path}) should have the same length as "
                f"the hierarchy ({self.dataset.hierarchy}) of {self}"
            )
        # Set a default ID of None for all parent frequencies that could be
        # inferred from a row at this depth
        # ids = {f: None for f in self.dataset.space}
        ids = dict(zip(self.dataset.hierarchy, tree_path))
        # Infer IDs and add them to those explicitly in the hierarchy
        ids.update(self.dataset.infer_ids(ids, metadata=metadata))
        # Calculate the combined freqs after each layer is added
        cummulative_freq = self.dataset.space(0)
        for layer_str in self.dataset.hierarchy:
            layer_freq = self.dataset.space[layer_str]
            # If all the axes introduced by the layer not present in parent layers
            # and none of the IDs of these axes have been inferred from other IDs,
            # then the ID of the axis out of the layer's axes with the least-
            # significant bit can be considered to be equivalent to the
            # ID of the layer and the IDs of the other axes of the layer set to None
            # (the order of # the bits in the DataSpace class should be arranged to
            # account for this default behaviour).
            #
            # For example, given a hierarchy of ['subject', 'session'] in the `Clinical`
            # data space, no groups are assumed to be present by default (i.e. if not
            # specified by the `id_patterns` attr of the dataset), and the `member`
            # ID is assumed to be equivalent to the `subject` ID, since `member`
            # correspdonds to the least significant bit in the value of the subject in
            # the `Clinical` data space enum.
            #
            # Conversely, the timepoint can't be assumed to be equal to the `session`
            # ID, since the session ID could be expected to also contain both the `member` and
            # `group` ID in it, and should be explicitly extracted by via `id_patterns`
            #
            #       session ID: MRH010_CONTROL03_MR02
            #
            # with the '02' part representing as the timepoint can be extracted with the
            #
            #       id_inference = {
            #           'timepoint': r'session:id:.*MR(0-9+)$'
            #       }
            layer_span = [str(f) for f in layer_freq.span()]
            if not (layer_freq & cummulative_freq) and not any(
                f in ids for f in layer_span
            ):
                for freq in layer_span[:-1]:
                    ids[freq] = None
                ids[layer_span[-1]] = ids[layer_str]
            cummulative_freq |= layer_freq
        assert cummulative_freq == self.dataset.space.leaf()
        missing_axes = set(str(f) for f in self.dataset.space.axes()) - set(ids)
        if missing_axes:
            logger.debug(
                "Leaf node at %s is missing explicit IDs for the following axes, %s"
                ", they will be set to None, noting that an error will be raised if there "
                " multiple nodes for this session. In that case,  set 'id-patterns' on the "
                "dataset to extract the missing axis IDs from composite IDs or row "
                "metadata",
                tree_path,
                missing_axes,
            )
            for m in missing_axes:
                ids[m] = None
        # # Set or override any inferred IDs within the ones that have been
        # # explicitly provided
        # clashing_ids = set(ids) & set(additional_ids)
        # if clashing_ids:
        #     raise ArcanaUsageError(
        #         f"Additional IDs clash with those inferred: {clashing_ids}"
        #     )
        # ids.update(additional_ids)
        # Create composite IDs for non-basis frequencies if they are not
        # explicitly in the layer dimensions
        for freq in set(self.dataset.space) - set(self.dataset.space.axes()):
            freq_str = str(freq)
            if freq_str not in ids:
                id = tuple(ids[str(b)] for b in freq.span() if ids[str(b)] is not None)
                if id:
                    if len(id) == 1:
                        id = id[0]
                    ids[freq_str] = id
        # Determine whether leaf node is included in the dataset definition according
        # to the include and exclude criteria
        exclusions = []
        for freq, include in self.dataset.include.items():
            freq_id = ids[freq]
            if (isinstance(include, list) and freq_id not in include) or (
                isinstance(include, str) and not re.match(include, freq_id)
            ):
                exclusions.append(freq)
                logger.debug(
                    f"skipping adding leaf at {tree_path} as {str(freq)} ID "
                    f"'{freq_id}' is not explicitly included: {include}"
                )
        for freq, exclude in self.dataset.exclude.items():
            freq_id = ids[freq]
            if (isinstance(exclude, list) and freq_id in exclude) or (
                isinstance(exclude, str) and re.match(exclude, freq_id)
            ):
                exclusions.append(freq)
                logger.debug(
                    f"skipping adding leaf at {tree_path} as {str(freq)} ID "
                    f"'{freq_id}' is explicitly excluded: {exclude}"
                )
        if not exclusions:
            row = self._add_row(
                ids={f: ids.get(str(f)) for f in self.dataset.space},
                row_frequency=self.dataset.space.leaf(),
            )
        else:
            row = None
        return row, exclusions

    def _add_row(self, ids: dict[DataSpace, str], row_frequency):
        """Adds a row to the dataset, creating all parent "aggregate" rows
        (e.g. for each subject, group or timepoint) where required

        Parameters
        ----------
        ids : dict[DataSpace, str]
            ids of the row in all frequencies that it intersects
        row: DataRow
            The row to add into the data tree

        Raises
        ------
        ArcanaDataTreeConstructionError
            If inserting a multiple IDs of the same class within the tree if
            one of their ids is None
        """
        # logger.debug(
        #     "Found %s row in %s dataset: %s", row_frequency, self.dataset_id, ids
        # )
        row_frequency = self.dataset.parse_frequency(row_frequency)
        row = DataRow(ids=ids, frequency=row_frequency, dataset=self.dataset)
        # Create new data row
        row_dict = self.root.children[row.frequency]
        if row.id in row_dict:
            raise ArcanaDataTreeConstructionError(
                f"ID clash ({row.id}) between rows inserted into data " "tree"
            )
        row_dict[row.id] = row
        # Insert root row
        # Insert parent rows if not already present and link them with
        # inserted row
        for parent_freq, parent_id in row.ids.items():
            if not parent_freq:
                continue  # Don't need to insert root row again
            diff_freq = (row.frequency ^ parent_freq) & row.frequency
            if diff_freq:
                # logger.debug(f'Linking parent {parent_freq}: {parent_id}')
                try:
                    parent_row = self.dataset.row(parent_freq, parent_id)
                except ArcanaNameError:
                    # logger.debug(
                    #     f'Parent {parent_freq}:{parent_id} not found, adding')
                    parent_ids = {
                        f: i
                        for f, i in row.ids.items()
                        if (f.is_parent(parent_freq) or f == parent_freq)
                    }
                    parent_row = self._add_row(parent_ids, parent_freq)
                # Set reference to level row in new row
                diff_id = row.frequency_id(diff_freq)
                children_dict = parent_row.children[row_frequency]
                if diff_id in children_dict:
                    raise ArcanaDataTreeConstructionError(
                        f"ID clash between rows inserted into data tree, {diff_id}, "
                        f"in {diff_freq} children of {parent_row} "
                        f"({children_dict[diff_id]} and {row}). You may "
                        f"need to set the `id_patterns` attr of the dataset "
                        "to disambiguate ID components (e.g. how to extract "
                        "the timepoint ID from a session label)"
                    )
                children_dict[diff_id] = row
        return row

    def _set_root(self):
        self.root = DataRow(
            ids={self.dataset.root_freq: None},
            frequency=self.dataset.root_freq,
            dataset=self.dataset,
        )



from __future__ import annotations
import logging
import re
from abc import abstractmethod, ABCMeta
from pathlib import Path
import attrs
import typing as ty
import yaml
from arcana.core.utils.serialize import (
    asdict,
    fromdict,
)
import arcana
from fileformats.core import DataType
from arcana.core.utils.misc import (
    get_config_file_path,
    NestedContext,
)
from arcana.core.utils.packaging import list_subclasses
from arcana.core.exceptions import (
    ArcanaUsageError,
    ArcanaNameError,
    ArcanaError,
    ArcanaDataTreeConstructionError,
)


DS = ty.TypeVar("DS", bound="DataStore")

logger = logging.getLogger("arcana")


if ty.TYPE_CHECKING:  # pragma: no cover
    from ..set import Dataset
    from ..tree import DataTree
    from ..entry import DataEntry
    from ..row import DataRow


@attrs.define
class ConnectionManager(NestedContext):

    store: ty.Any = None
    session: ty.Any = attrs.field(default=None, init=False)

    def __getattr__(self, attr_name):
        return getattr(self.session, attr_name)

    def enter(self):
        self.session = self.store.connect()

    def exit(self):
        self.store.disconnect(self.session)
        self.session = None


@attrs.define
class DataStore(metaclass=ABCMeta):
    """
    Abstract base class for all data store adapters. A data store can be an external
    data management system, e.g. XNAT, OpenNeuro, Datalad or just a defined structure
    of how to lay out data within a file-system, e.g. BIDS.

    For a data management system/data structure to be compatible with Arcana, it must
    meet a number of criteria. In Arcana, a store is assumed to

        * contain multiple projects/datasets addressable by unique IDs.
        * organise data within each project/dataset in trees
        * store arbitrary numbers of data "items" (e.g. "file-sets" and fields) within
          each tree node (including non-leaf nodes) addressable by unique "paths" relative
          to the node.
        * allow derivative data to be stored within in separate namespaces for different
          analyses on the same data
    """

    # name: str = None
    connection: ConnectionManager = attrs.field(
        factory=ConnectionManager, init=False, hash=False, repr=False, eq=False
    )

    def __attrs_post_init__(self):
        self.connection.store = self

    CONFIG_NAME = "stores"
    SUBPACKAGE = "data"
    VERSION_KEY = "store-version"
    VERSION = "1.0.0"
    # alternative name used to save datasets that are named "" in cases where "" is
    # not appropriate
    EMPTY_DATASET_NAME = "_"

    ##############
    # Public API #
    ##############

    def save(self, name: str = None, config_path: Path = None):
        """Saves the configuration of a DataStore in 'stores.yaml'

        Parameters
        ----------
        name
            The name under which to save the data store
        config_path : Path, optional
            the path to save the config file to, defaults to `~/.arcana/stores.yaml`
        """
        if name in self.singletons():
            raise ArcanaNameError(
                name, f"Name '{name}' clashes with built-in type of store"
            )
        if name is None:
            if self.name is None:
                raise ArcanaNameError(
                    f"Must provide name to save store {self} as as it doesn't have one "
                    "already"
                )
        else:
            self.name = name
        entries = self.load_saved_configs()
        # connect to store in case it is needed in the asdict method and to
        # test the connection in general before it is saved
        dct = self.asdict()
        with self.connection:
            entries[dct.pop("name")] = dct
        self.save_configs(entries, config_path=config_path)

    def asdict(self, **kwargs):
        return asdict(self, **kwargs)

    @classmethod
    def load(
        cls: DataStore, name: str, config_path: Path = None, **kwargs
    ) -> DataStore:
        """Loads a DataStore from that has been saved in the configuration file.
        If no entry is saved under that name, then it searches for DataStore
        sub-classes with aliases matching `name` and checks whether they can
        be initialised without any parameters.

        Parameters
        ----------
        name : str
            Name that the store was saved under
        config_path : Path, optional
            path to the config file, defaults to `~/.arcana/stores.yaml`
        **kwargs
            keyword args passed to the store, overriding values stored in the
            entry

        Returns
        -------
        DataStore
            The data store retrieved from the stores.yaml file

        Raises
        ------
        ArcanaNameError
            If the name is not found in the saved stores
        """
        entries = cls.load_saved_configs(config_path)
        try:
            entry = entries[name]
        except KeyError:
            try:
                return cls.singletons()[name]
            except KeyError:
                raise ArcanaNameError(
                    name, f"No saved data store or built-in type matches '{name}'"
                )
        else:
            entry.update({k: v for k, v in kwargs.items() if v is not None})
            entry["name"] = name
            store = fromdict(entry)  # Would be good to use a class resolver here
        return store

    @classmethod
    def remove(cls, name: str, config_path: Path = None):
        """Removes the entry saved under 'name' in the config file

        Parameters
        ----------
        name
            Name of the configuration to remove
        """
        entries = cls.load_saved_configs(config_path)
        del entries[name]
        cls.save_configs(entries)

    def define_dataset(
        self, id, space=None, hierarchy=None, id_patterns=None, **kwargs
    ) -> Dataset:
        """
        Creates a Arcana dataset definition for an existing data in the
        data store.

        Parameters
        ----------
        id : str
            The ID (or file-system path) of the project (or directory) within
            the store
        space: DataSpace
            The data space of the dataset
        hierarchy: list[str]
            The hierarchy of the dataset
        id_patterns : dict[str, str], optional
            Patterns used to infer row IDs not explicitly within the hierarchy of the
            data tree, e.g. groups and timepoints in an XNAT project with subject>session
            hierarchy
        space : EnumMeta
            The DataSpace enum that defines the frequencies (e.g.
            per-session, per-subject,...) present in the dataset.
        **kwargs:
            Keyword args passed on to the Dataset init method

        Returns
        -------
        Dataset
            the newly defined dataset
        """
        if space is None:
            try:
                space = self.DEFAULT_SPACE
            except AttributeError as e:
                raise ArcanaUsageError(
                    "'space' kwarg must be specified for datasets in "
                    f"{type(self)} stores"
                ) from e
        if hierarchy is None:
            try:
                hierarchy = list(self.DEFAULT_HIERARCHY)
            except AttributeError:
                hierarchy = [str(max(space))]  # one-layer with only leaf nodes
        # if id_patterns is None:
        #     try:
        #         id_patterns = dict(self.DEFAULT_ID_PATTERNS)
        #     except AttributeError:
        #         pass
        from arcana.core.data.set import (
            Dataset,
        )  # avoid circular imports it is imported here rather than at the top of the file

        dataset = Dataset(
            id,
            store=self,
            space=space,
            hierarchy=hierarchy,
            id_patterns=id_patterns,
            **kwargs,
        )
        return dataset

    def save_dataset(self, dataset: Dataset, name: str = ""):
        """Save metadata in project definition file for future reference

        Parameters
        ----------
        dataset : Dataset
            the dataset to save
        name : str, optional
            the name for the definition to distinguish from other definitions on
            the same data, by default None
        """
        if name is None:
            name = ""
        save_name = name if name else self.EMPTY_DATASET_NAME
        definition = asdict(dataset, omit=["store", "name"])
        definition[self.VERSION_KEY] = self.VERSION
        if name is None:
            name = dataset.name
        with self.connection:
            self.save_dataset_definition(dataset.id, definition, name=save_name)

    def load_dataset(self, id, name: str = "", **kwargs) -> Dataset:
        """Load an existing dataset definition

        Parameters
        ----------
        id : str
            ID of the dataset within the store
        name : str, optional
            name of the dataset definition, which distinguishes it from alternative
            definitions on the same data, by default None

        Returns
        -------
        Dataset
            the loaded dataset

        Raises
        ------
        KeyError
            if the dataset is not found
        """
        if name is None:
            name = ""
        saved_name = name if name else self.EMPTY_DATASET_NAME
        with self.connection:
            dct = self.load_dataset_definition(id, saved_name)
        if dct is None:
            raise KeyError(f"Did not find a dataset '{id}@{name}'")
        store_version = dct.pop(self.VERSION_KEY)
        self.check_store_version(store_version)
        return fromdict(dct, id=id, name=name, store=self, **kwargs)

    def create_dataset(
        self,
        id: str,
        leaves: list[tuple[str, ...]],
        hierarchy: list[str],
        space: type,
        name: str = None,
        id_patterns: dict[str, str] = None,
        **kwargs,
    ) -> Dataset:
        """Creates a new dataset with new rows to store data in

        Parameters
        ----------
        id : str
            ID of the dataset
        leaves : list[tuple[str, ...]]
            the list of tuple IDs (at each level of the tree)
        name : str, optional
            name of the dataset, if provided the dataset definition will be saved. To
            save the dataset with the default name pass an empty string.
        hierarchy : list[str], optional
            hierarchy of the dataset tree
        space : type, optional
            the space of the dataset
        id_patterns : dict[str, str]
            Patterns for inferring IDs of rows not explicitly present in the hierarchy of
            the data tree. See ``DataStore.infer_ids()`` for syntax

        Returns
        -------
        Dataset
            the newly created dataset
        """
        self.create_data_tree(
            id=id,
            leaves=leaves,
            hierarchy=hierarchy,
            space=space,
        )
        dataset = self.define_dataset(
            id=id,
            hierarchy=hierarchy,
            space=space,
            id_patterns=id_patterns,
            **kwargs,
        )
        if name is not None:
            dataset.save(name=name)
        return dataset

    def import_dataset(
        self,
        id: str,
        dataset: Dataset,
        column_names: list[str],
        id_patterns: dict[str, str] = None,
        use_original_paths: bool = False,
        **kwargs,
    ):
        """Import a dataset from another store, transferring metadata and columns
        defined on the original dataset

        Parameters
        ----------
        id : str
            the ID of the dataset within this store
        dataset : Dataset
            the dataset to import
        column_names : list[str]
            list of columns to be included in the imported dataset
        id_patterns : dict[str, str]
            Patterns for inferring IDs of rows not explicitly present in the hierarchy of
            the data tree. See ``DataStore.infer_ids()`` for syntax
        use_original_paths : bool, optional
            use the original paths in the source store instead of renaming the imported
            entries to match their column names
        **kwargs:
            keyword arguments passed through to the `create_data_tree` method
        """
        if use_original_paths:
            raise NotImplementedError
        # Create a new dataset in the store to import the data into
        imported = self.create_dataset(
            id,
            space=dataset.space,
            hierarchy=dataset.hierarchy,
            leaves=dataset.row_ids(),
            id_patterns=id_patterns,
            metadata=dataset.metadata,
            **kwargs,
        )
        # Loop through columns
        if column_names is None:
            column_names = list(dataset.columns)
        for column_name in column_names:
            column = dataset.columns[column_name]
            path = column.name if not column.is_sink else column.path
            # Create columns in imported dataset
            imported_col = imported.add_sink(
                name=column.name,
                datatype=column.datatype,
                path=path,
                row_frequency=column.row_frequency,
            )
            # Copy across data from dataset to import
            for cell in column.cells():
                imported_col[cell.row.id] = cell.item
        imported.save(name="")

    @classmethod
    def singletons(cls):
        """Returns stores in a dictionary indexed by their aliases, for which there
        only needs to be a single instance"""
        try:
            return cls._singletons
        except AttributeError:
            pass
        # If not saved in the configuration file search for sub-classes
        # whose alias matches `name` and can be initialised without params
        cls._singletons = {}
        for store_cls in list_subclasses(arcana, DataStore):
            try:
                store = store_cls()
            except Exception:
                pass
            else:
                cls._singletons[store.name] = store
        return cls._singletons

    @classmethod
    def load_saved_configs(cls, config_path: Path = None) -> dict[str, ty.Any]:
        """Loads the saved data store configurations from the the user's home
        directory

        Parameters
        ----------
        config_path : Path, optional
            the file-system path to the configuration, by default uses one in ~/.arcana

        Returns
        -------
        dict[str, ty.Any]
            dictionary containing the saved configs
        """
        if config_path is None:
            config_path = get_config_file_path(cls.CONFIG_NAME)
        if config_path.exists():
            with open(config_path) as f:
                configs = yaml.load(f, Loader=yaml.Loader)
        else:
            configs = {}
        return configs

    @classmethod
    def save_configs(cls, configs: dict[str, ty.Any], config_path: Path = None):
        """_summary_

        Parameters
        ----------
        configs : dict[str, ty.Any]
            dictionary containing the configs to save
        config_path : Path, optional
            the file-system path to the configuration, by default uses one in ~/.arcana
        """
        if config_path is None:
            config_path = get_config_file_path(cls.CONFIG_NAME)
        with open(config_path, "w") as f:
            yaml.dump(configs, f)

    @classmethod
    def infer_ids(
        cls,
        ids: dict[str, str],
        id_patterns: dict[str, str],
        metadata: dict[str, dict[str, str]] = None,
    ):
        """Infer IDs from those explicitly provided by using the inference patterns
        provided to the dataset definition.

        Not all IDs will appear explicitly within the hierarchy of the data
        tree, and some will need to be inferred by extracting components of
        more composite labels or associated metadata.

        For example, given a set of subject IDs that combination of the ID of
        the group that they belong to and the member ID within that group
        (i.e. matched test & control would have same member ID)

            CONTROL01, CONTROL02, CONTROL03, ... and TEST01, TEST02, TEST03

        the group ID can be extracted by providing a dictionary of the IDs to
        be inferred and the patterns to derive them from. Extracted components are
        specified by 2 or 3-tuples delimited by ':', where

            1. frequency of the layer to extract the component from.
            2. the metadata value to extract. "id" is a reserved key, which refers to the layer label
            3. a regular expression with a single capturing group

        ::

            id_patterns = {
                'group': r"subject:id:([A-Z]+)[0-9]+',
                'member': r"subject:id:[A-Z]+([0-9]+)',
            }

        If you want to compose an ID out of one or more extracted patterns, they can
        be enclosed by ``#`` symbols::

            id_patterns = {
                "timepoint": r"T#session:order#"
            }

        where "order" is a special metadata field added by the data store designating
        the order in which the session was acquired within the subject. This pattern will
        produce timepoint IDs "T1", "T2", "T3", ...

        Parameters
        ----------
        ids : dict[str, str]
            explicitly provided IDs
        id_patterns : dict[str, str]
            patterns used to infer IDs not explicitly in the hierarchy of the dataset
        metadata : dict[str, dict[str, str]]
            metadata associated with the nodes in each layer. Can be used as an input
            to a pattern

        Return
        ------
        inferred_ids : dict[str, str]
            IDs inferred from the decomposition
        """
        if metadata is None:
            metadata = {}
        inferred_ids = {}
        if id_patterns is not None:
            conflicting = set(ids) & set(id_patterns)
            if conflicting:
                raise ArcanaDataTreeConstructionError(
                    "Inferred IDs from decomposition conflict with explicitly provided IDs: "
                    + str(conflicting)
                )
            for freq, pattern in id_patterns.items():
                comps = cls.pattern_comp_re.findall(pattern)
                if not comps:
                    comps = [pattern]
                    whole_str = True
                else:
                    whole_str = False
                substitutions = []
                for comp in comps:
                    parts = comp.strip("#").split(":")
                    source_freq = parts[0] if parts[0] else freq
                    attr_name = parts[1] if len(parts) >= 2 and parts[1] else "ID"
                    regex = ":".join(parts[2:])
                    if attr_name.lower() == "id":
                        attr = ids[source_freq]
                        attr_name = attr_name.upper()
                    else:
                        try:
                            attr = str(metadata[source_freq][attr_name])
                        except KeyError:
                            raise ArcanaDataTreeConstructionError(
                                f"'{ids[source_freq]}' {source_freq} row doesn't have "
                                f"the metadata field '{attr_name}'"
                            )
                    if regex:
                        match = re.fullmatch(regex, attr)
                        if not match or len(match.groups()) != 1:
                            match_msg = (
                                f"matched {len(match.groups())} groups"
                                if match
                                else "didn't match the pattern"
                            )
                            raise ArcanaDataTreeConstructionError(
                                f"Provided ID-pattern component,'{regex}', needs to match "
                                f"exactly one group on '{attr_name}' attribute of "
                                f"'{ids[source_freq]}' {source_freq} row, '{attr}', when it "
                                + match_msg
                            )
                        attr = match.group(1)
                    substitutions.append(attr)
                if whole_str:
                    assert len(substitutions) == 1
                    inferred_id = substitutions[0]
                else:
                    inferred_id = pattern
                    for sub in substitutions:
                        inferred_id = cls.pattern_comp_re.subn(
                            sub, inferred_id, count=1
                        )[0]
                inferred_ids[freq] = inferred_id
        return inferred_ids

    ####################
    # Abstract methods #
    ####################

    @abstractmethod
    def populate_tree(self, tree: DataTree):
        """
        Populates the nodes of the data tree with those found in the dataset using
        the ``DataTree.add_leaf`` method for every "leaf" node of the dataset tree.

        Parameters
        ----------
        tree : DataTree
            The tree to populate with nodes
        """

    @abstractmethod
    def populate_row(self, row: DataRow):
        """
        Populate a row with all data entries found in the corresponding node in the data
        store (e.g. files within a directory, scans within an XNAT session) using the
        ``DataRow.add_entry`` method. Within a node/row there are assumed to be two types
        of entries, "primary" entries (e.g. acquired scans) common to all analyses performed
        on the dataset and "derivative" entries corresponding to intermediate outputs
        of previously performed analyses. These types should be stored in separate
        namespaces so there is no chance of a derivative overriding a primary data item.

        The name of the dataset/analysis a derivative was generated by is appended to
        to a base path, delimited by "@", e.g. "brain_mask@my_analysis". The dataset
        name is left blank by default, in which case "@" is just appended to the
        derivative path, i.e. "brain_mask@".

        Parameters
        ----------
        row : DataRow
            The row to populate with entries
        """

    @abstractmethod
    def get(self, entry: DataEntry, datatype: type) -> DataType:
        """
        Gets the data item corresponding to the given entry

        Parameters
        ----------
        entry : DataEntry
            the data entry to update
        datatype : type
            the datatype to interpret the entry's item as

        Returns
        -------
        item : DataType
            the item stored within the specified entry
        """

    @abstractmethod
    def put(self, item: DataType, entry: DataEntry) -> DataType:
        """
        Updates the item in the data store corresponding to the given data entry

        Parameters
        ----------
        item : DataType
            the item to replace the current item in the data store
        entry: DataEntry
            the data entry to update

        Returns
        -------
        cached : DataType
            returns the cached version of the item, if applicable
        """

    @abstractmethod
    def put_provenance(self, provenance: dict[str, ty.Any], entry: DataEntry):
        """Stores provenance information for a given data item in the store

        Parameters
        ----------
        entry: DataEntry
            The item to store the provenance data for
        provenance: dict[str, Any]
            The provenance data to store
        """

    @abstractmethod
    def get_provenance(self, entry: DataEntry) -> dict[str, ty.Any]:
        """Stores provenance information for a given data item in the store

        Parameters
        ----------
        entry: DataEntry
            The item to store the provenance data for

        Returns
        -------
        provenance: dict[str, Any] or None
            The provenance data stored in the repository for the data item.
            None if no provenance data has been stored
        """

    @abstractmethod
    def save_dataset_definition(
        self, dataset_id: str, definition: ty.Dict[str, ty.Any], name: str
    ):
        """Save definition of dataset within the store

        Parameters
        ----------
        dataset_id: str
            The ID/path of the dataset within the store
        definition: dict[str, Any]
            A dictionary containing the dct Dataset to be saved. The
            dictionary is in a format ready to be dumped to file as JSON or
            YAML.
        name: str
            Name for the dataset definition to distinguish it from other
            definitions for the same directory/project
        """

    @abstractmethod
    def load_dataset_definition(
        self, dataset_id: str, name: str
    ) -> ty.Dict[str, ty.Any]:
        """Load definition of a dataset saved within the store

        Parameters
        ----------
        dataset_id: str
            The ID (e.g. file-system path, XNAT project ID) of the project
        name: str
            Name for the dataset definition to distinguish it from other
            definitions for the same directory/project

        Returns
        -------
        definition: dict[str, Any]
            A dct Dataset object that was saved in the data store
        """

    @abstractmethod
    def connect(self) -> ty.Any:
        """
        If a connection session is required to the store manage it here

        Returns
        ----------
        session : Any
            a session object that will be stored in the connection manager and
            accessible at `DataStore.connection`
        """

    @abstractmethod
    def disconnect(self, session: ty.Any):
        """
        If a connection session is required to the store manage it here

        Parameters
        ----------
        session : Any
            the session object returned by `connect` to be closed gracefully
        """

    @abstractmethod
    def site_licenses_dataset(self):
        """Can be overridden by subclasses to provide a dataset to hold site-wide licenses"""

    @abstractmethod
    def create_data_tree(
        self,
        id: str,
        leaves: list[tuple[str, ...]],
        hierarchy: list[str],
        space: type,
        **kwargs,
    ):
        """Creates a new empty dataset within in the store. Used in test routines and
        importing/exporting datasets between stores

        Parameters
        ----------
        id : str
            ID for the newly created dataset
        leaves : list[tuple[str, ...]]
                        list of IDs for each leaf node to be added to the dataset. The IDs for each
            leaf should be a tuple with an ID for each level in the tree's hierarchy, e.g.
            for a hierarchy of [subject, timepoint] ->
            [("SUBJ01", "TIMEPOINT01"), ("SUBJ01", "TIMEPOINT02"), ....]
        hierarchy: list[str]
            the hierarchy of the dataset to be created
        space : type(DataSpace)
            the data space of the dataset
        id_patterns : dict[str, str]
            Patterns for inferring IDs of rows not explicitly present in the hierarchy of
            the data tree. See ``DataStore.infer_ids()`` for syntax
        **kwargs
            implementing methods should take wildcard **kwargs to allow compatibility
            with future arguments that might be added
        """
        raise NotImplementedError

    @abstractmethod
    def create_entry(self, path: str, datatype: type, row: DataRow) -> DataEntry:
        """Creates an "entry" in the store to hold a new data item

        Parameters
        ----------
        path : str
            path to the entry relative to the data "row"
        datatype : type
            the datatype of the entry
        row : DataRow
            the row (tree node) to create the entry in

        Returns
        -------
        entry : DataEntry
            the newly created entry
        """

    # Can be overridden if necessary (e.g. the underlying store only returns new URI
    # when a new item is added)
    def post(
        self, item: DataType, path: str, datatype: type, row: DataRow
    ) -> DataEntry:
        """Inserts the item within a newly created entry in the data store

        Parameters
        ----------
        item : DataType
            the item to insert
        path : str
            the path to the entry relative to the data row
        datatype : type
            the datatype of the entry
        row : DataRow
            the data row to insert the entry into

        Returns
        -------
        entry : DataEntry
            the inserted entry
        """
        with self.connection:
            entry = self.create_entry(path, datatype, row)
            self.put(item, entry)

    ##################
    # Helper methods #
    ##################

    def check_store_version(self, store_version: str):
        """Check whether version store used to save the dataset is compatible with the
        current version of the software. Can be overridden by store subclasses where
        appropriate

        Parameters
        ----------
        store_version : str
            version of the store used to save the dataset

        Raises
        ------
        ArcanaError
            if the saved version isn't compatible
        """
        if store_version != self.VERSION:
            raise ArcanaError(
                f"Stored version of dataset ({store_version}) does not match current "
                f"version of {type(self).__name__} ({self.VERSION})"
            )

    pattern_comp_re = re.compile(r"#[^\#]+#")



from pathlib import Path
import pytest
from arcana.core.data.set.base import Dataset
from arcana.core.analysis.salience import ColumnSalience
from arcana.core.data.quality import DataQuality
from arcana.testing import TestDataSpace
from arcana.core.cli.dataset import define, add_source, add_sink, missing_items, export
from fileformats.text import Plain as Text
from fileformats.generic import Directory
from arcana.core.utils.misc import show_cli_trace
from arcana.testing.data.blueprint import TEST_DATASET_BLUEPRINTS
from arcana.testing.data import MockRemote
from arcana.dirtree import DirTree


ARBITRARY_INTS_A = [234221, 93380, 43271, 137483, 30009, 214205, 363526]
ARBITRARY_INTS_B = [353726, 29202, 32867, 129872, 12281, 776524, 908763]


def get_arbitrary_slice(i, dim_length):
    a = ARBITRARY_INTS_A[i] % dim_length
    b = ARBITRARY_INTS_B[i] % dim_length
    lower = min(a, b)
    upper = max(a, b) + 1
    return lower, upper


def test_add_column_cli(saved_dataset: Dataset, cli_runner):
    # Get CLI name for dataset (i.e. file system path prepended by 'dirtree//')
    # Add source to loaded dataset
    saved_dataset.add_source(
        name="a_source",
        path="file1",
        datatype=Text,
        row_frequency=TestDataSpace.d,
        quality_threshold=DataQuality.questionable,
        order=1,
        required_metadata={},
        is_regex=False,
    )
    # Add source column to saved dataset
    result = cli_runner(
        add_source,
        [
            saved_dataset.locator,
            "a_source",
            "fileformats.text:Plain",
            "--path",
            "file1",
            "--row-frequency",
            "d",
            "--quality",
            "questionable",
            "--order",
            "1",
            "--no-regex",
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    # Add source to loaded dataset
    saved_dataset.add_sink(
        name="a_sink",
        path="deriv",
        datatype=Text,
        row_frequency=TestDataSpace.d,
        salience=ColumnSalience.qa,
    )
    result = cli_runner(
        add_sink,
        [
            saved_dataset.locator,
            "a_sink",
            "fileformats.text:Plain",
            "--path",
            "deriv",
            "--row-frequency",
            "d",
            "--salience",
            "qa",
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    # Reload the saved dataset and check the parameters were saved/loaded
    # correctly
    loaded_dataset = Dataset.load(saved_dataset.locator)
    assert saved_dataset.columns == loaded_dataset.columns


@pytest.mark.skip("Not implemented")
def test_add_missing_items_cli(saved_dataset: Dataset, cli_runner):
    result = cli_runner(missing_items, [])
    assert result.exit_code == 0, show_cli_trace(result)


def test_define_cli(dataset: Dataset, cli_runner):
    blueprint = dataset.__annotations__["blueprint"]
    # Get CLI name for dataset (i.e. file system path prepended by 'file//')
    path = dataset.locator
    # Start generating the arguments for the CLI
    args = list(blueprint.hierarchy)
    # Generate "arbitrary" values for included and excluded from dim length
    # and index
    included = {}
    excluded = {}
    for i, (dim_length, axis) in enumerate(zip(blueprint.dim_lengths, dataset.space)):
        a, b = get_arbitrary_slice(i, dim_length)
        if i % 2:
            included[str(axis)] = f"{a}:{b}"
        elif (b - a) < dim_length:  # Check that we aren't excluding all
            excluded[str(axis)] = f"{a}:{b}"
    # Add include and exclude options
    for axis, slce in included.items():
        args.extend(["--include", axis, slce])
    for axis, slce in excluded.items():
        args.extend(["--exclude", axis, slce])
    args.extend(["--space", "arcana.testing:TestDataSpace"])
    # Run the command line
    result = cli_runner(define, [path, *args])
    # Check tool completed successfully
    assert result.exit_code == 0, show_cli_trace(result)
    # Reload the saved dataset and check the parameters were saved/loaded
    # correctly
    loaded_dataset = Dataset.load(path)
    assert loaded_dataset.hierarchy == blueprint.hierarchy
    assert loaded_dataset.include == included
    assert loaded_dataset.exclude == excluded


def test_export_import_roundtrip(cli_runner, work_dir: Path, arcana_home):
    blueprint = TEST_DATASET_BLUEPRINTS["skip_single"]
    cache_dir = work_dir / "remote-cache"
    cache_dir.mkdir()
    remote_dir = work_dir / "remote-dir"
    remote_dir.mkdir()
    mock_remote = MockRemote(
        server="https://a.server.com",
        user="admin",
        password="admin",
        cache_dir=work_dir / "remote-cache",
        remote_dir=remote_dir,
    )
    mock_remote.save("my_mock_remote")
    original = blueprint.make_dataset(store=mock_remote, dataset_id="original")
    original.add_sink(
        name="doubledir1",
        datatype=Directory,
        path="doubledir1",
    )
    original.add_sink(
        name="doubledir2",
        datatype=Directory,
        path="doubledir2",
    )
    original.save()
    local_dataset_id = str(work_dir / "exported")
    # Add source column to saved dataset
    result = cli_runner(
        export,
        [
            original.locator,
            "dirtree",
            local_dataset_id,
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    local_dataset = DirTree().load_dataset(local_dataset_id)
    result = cli_runner(
        export,
        [
            local_dataset.locator,
            "my_mock_remote",
            "reimported",
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    reimported = mock_remote.load_dataset("reimported")
    reimported.id = "original"
    assert reimported == original



from __future__ import annotations
from operator import itemgetter
import pytest
from fileformats.text import Plain as PlainText
from arcana.core.exceptions import ArcanaUsageError, ArcanaDataTreeConstructionError
from arcana.dirtree import DirTree
from arcana.testing.data.blueprint import TestDatasetBlueprint, FileSetEntryBlueprint
from arcana.testing.data.space import TestDataSpace


TEST_INCLUSIONS = {
    "all": (
        TestDatasetBlueprint(  # dataset name
            space=TestDataSpace,
            hierarchy=["a", "b", "c", "abcd"],
            dim_lengths=[1, 2, 3, 4],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=PlainText, filenames=["file1.txt"]
                ),
            ],
            id_patterns={"d": r"abcd::.*(d\d+)"},
        ),
        [
            "a0b0c0d0",
            "a0b0c0d1",
            "a0b0c0d2",
            "a0b0c0d3",
            "a0b0c1d0",
            "a0b0c1d1",
            "a0b0c1d2",
            "a0b0c1d3",
            "a0b0c2d0",
            "a0b0c2d1",
            "a0b0c2d2",
            "a0b0c2d3",
            "a0b1c0d0",
            "a0b1c0d1",
            "a0b1c0d2",
            "a0b1c0d3",
            "a0b1c1d0",
            "a0b1c1d1",
            "a0b1c1d2",
            "a0b1c1d3",
            "a0b1c2d0",
            "a0b1c2d1",
            "a0b1c2d2",
            "a0b1c2d3",
        ],
    ),
    "include": (
        TestDatasetBlueprint(  # dataset name
            space=TestDataSpace,
            hierarchy=["a", "b", "c", "abcd"],
            dim_lengths=[1, 2, 3, 4],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=PlainText, filenames=["file1.txt"]
                ),
            ],
            include={
                "c": ["c1", "c2"],
                "a": "a0",
                "d": ["d1", "d2", "d3"],
            },
            id_patterns={"d": r"abcd::.*(d\d+)"},
        ),
        [
            "a0b0c1d1",
            "a0b0c1d2",
            "a0b0c1d3",
            "a0b0c2d1",
            "a0b0c2d2",
            "a0b0c2d3",
            "a0b1c1d1",
            "a0b1c1d2",
            "a0b1c1d3",
            "a0b1c2d1",
            "a0b1c2d2",
            "a0b1c2d3",
        ],
    ),
    "exclude": (
        TestDatasetBlueprint(  # dataset name
            space=TestDataSpace,
            hierarchy=["a", "b", "c", "abcd"],
            dim_lengths=[1, 2, 3, 4],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=PlainText, filenames=["file1.txt"]
                ),
            ],
            exclude={
                "c": ["c2"],
                "abcd": ["a0b0c1d1", "a0b1c1d2"],
                "d": ["d3"],
            },
            id_patterns={"d": r"abcd::.*(d\d+)"},
        ),
        [
            "a0b0c0d0",
            "a0b0c0d1",
            "a0b0c0d2",
            "a0b0c1d0",
            "a0b0c1d2",
            "a0b1c0d0",
            "a0b1c0d1",
            "a0b1c0d2",
            "a0b1c1d0",
            "a0b1c1d1",
        ],
    ),
    "regex": (
        TestDatasetBlueprint(  # dataset name
            space=TestDataSpace,
            hierarchy=["a", "b", "c", "abcd"],
            dim_lengths=[1, 2, 3, 4],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=PlainText, filenames=["file1.txt"]
                ),
            ],
            id_patterns={"bc": r"BC#b:id##c:id#"},
            include={
                "abcd": r"a\d+b\dc(\d+)d\1",
            },
            exclude={"bc": r"BC(\d)\1"},
        ),
        [
            "a0b0c0d0",
            "a0b0c1d1",
            "a0b0c2d2",
            "a0b1c0d0",
            "a0b1c1d1",
            "a0b1c2d2",
        ],
    ),
}


@pytest.mark.parametrize("fixture", TEST_INCLUSIONS.items(), ids=itemgetter(0))
def test_dataset_inclusion(
    fixture: tuple[str, tuple[TestDatasetBlueprint, list[str]]], work_dir
):
    test_name, (blueprint, expected) = fixture
    dataset_path = work_dir / test_name
    dataset = blueprint.make_dataset(store=DirTree(), dataset_id=dataset_path)
    assert sorted(dataset.row_ids()) == expected


def test_include_exclude_fail(work_dir):

    blueprint = TestDatasetBlueprint(  # dataset name
        space=TestDataSpace,
        hierarchy=["a", "b", "c", "abcd"],
        dim_lengths=[1, 2, 3, 4],
        entries=[
            FileSetEntryBlueprint(
                path="file1", datatype=PlainText, filenames=["file1.txt"]
            ),
        ],
        include={"session": r"a0.*d1.*"},
        exclude={"session": r"a0.*d2.*"},
    )
    with pytest.raises(
        ArcanaUsageError, match="Cannot provide both 'include' and 'exclude' arguments"
    ):
        blueprint.make_dataset(
            store=DirTree(), dataset_id=work_dir / "include-exclude-fail"
        )


def test_requires_id_pattern(work_dir):

    blueprint = TestDatasetBlueprint(  # dataset name
        space=TestDataSpace,
        hierarchy=["a", "b", "c", "abcd"],
        dim_lengths=[1, 2, 3, 4],
        entries=[
            FileSetEntryBlueprint(
                path="file1", datatype=PlainText, filenames=["file1.txt"]
            ),
        ],
    )
    with pytest.raises(
        ArcanaDataTreeConstructionError,
        match="ID clash between rows inserted into data tree",
    ):
        blueprint.make_dataset(
            store=DirTree(), dataset_id=work_dir / "include-exclude-fail"
        )



from arcana.core.cli.store import add, ls, remove, rename
from arcana.core.utils.misc import show_cli_trace
from arcana.core.data.store import DataStore

STORE_URI = "http://dummy.uri"
STORE_USER = "a_user"
STORE_PASSWORD = "a-password"


def test_store_cli(cli_runner, arcana_home, work_dir):
    store_name = "test-mock-remote"
    # Add new XNAT configuration
    result = cli_runner(
        add,
        [
            store_name,
            "testing:MockRemote",
            "--server",
            STORE_URI,
            "--user",
            STORE_USER,
            "--password",
            STORE_PASSWORD,
            "--option",
            "remote_dir",
            str(work_dir / "remote-dir"),
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    # List all saved and built-in stores
    result = cli_runner(ls, [])
    assert result.exit_code == 0, show_cli_trace(result)
    assert f"{store_name} - arcana.testing.data.store:MockRemote" in result.output
    assert "    server: " + STORE_URI in result.output


def test_store_cli_remove(cli_runner, arcana_home, work_dir):
    new_store_name = "a-new-mock"
    # Add new XNAT configuration
    cli_runner(
        add,
        [
            new_store_name,
            "testing:MockRemote",
            "--server",
            STORE_URI,
            "--user",
            STORE_USER,
            "--password",
            STORE_PASSWORD,
            "--option",
            "remote_dir",
            str(work_dir / "remote-dir"),
        ],
    )
    # Check store is saved
    result = cli_runner(ls, [])
    assert new_store_name in result.output

    cli_runner(remove, new_store_name)
    # Check store is gone
    result = cli_runner(ls, [])
    assert new_store_name not in result.output


def test_store_cli_rename(cli_runner, arcana_home, work_dir):
    old_store_name = "i123"
    new_store_name = "y456"
    # Add new XNAT configuration
    cli_runner(
        add,
        [
            old_store_name,
            "testing:MockRemote",
            "--server",
            STORE_URI,
            "--user",
            STORE_USER,
            "--password",
            STORE_PASSWORD,
            "--option",
            "remote_dir",
            str(work_dir / "remote-dir"),
        ],
    )
    # Check store is saved
    result = cli_runner(ls, [])
    assert "i123 - arcana.testing.data.store:MockRemote" in result.output

    cli_runner(rename, [old_store_name, new_store_name])
    # Check store is renamed
    result = cli_runner(ls, [])
    assert "i123 - arcana.testing.data.store:MockRemote" not in result.output
    assert "y456 - arcana.testing.data.store:MockRemote" in result.output


def test_store_cli_encrypt_credentials(cli_runner, arcana_home, work_dir):
    store_name = "another-test-mock"
    # Add new XNAT configuration
    result = cli_runner(
        add,
        [
            store_name,
            "testing:MockRemote",
            "--server",
            STORE_URI,
            "--user",
            STORE_USER,
            "--password",
            STORE_PASSWORD,
            "--option",
            "remote_dir",
            str(work_dir / "remote-dir"),
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    # Check credentials have been encrypted
    loaded_store = DataStore.load(store_name)
    assert loaded_store.password != ""
    assert loaded_store.password is not STORE_PASSWORD
    assert loaded_store.user != ""
    assert loaded_store.user is not STORE_USER



from __future__ import annotations
from abc import abstractmethod, ABCMeta
import re
import typing as ty
import logging
import attrs
from operator import attrgetter
from attrs.converters import optional
from fileformats.core import DataType
from fileformats.core.exceptions import FormatMismatchError
from arcana.core.exceptions import ArcanaDataMatchError
from ..analysis.salience import ColumnSalience
from .quality import DataQuality
from .space import DataSpace
from .cell import DataCell
from .entry import DataEntry

if ty.TYPE_CHECKING:  # pragma: no cover
    from .row import DataRow
    from .set.base import Dataset


logger = logging.getLogger("arcana")


@attrs.define(kw_only=True)
class DataColumn(metaclass=ABCMeta):

    name: str
    datatype: type = attrs.field()
    row_frequency: DataSpace = attrs.field(
        validator=attrs.validators.instance_of(DataSpace)
    )
    path: str = None
    dataset: Dataset = attrs.field(
        default=None, metadata={"asdict": False}, eq=False, hash=False, repr=False
    )
    _mismatch_log: list = attrs.field(
        default=None, eq=False, hash=False, repr=False, init=False
    )

    is_sink = False

    @datatype.validator
    def datatype_validator(self, _, datatype):
        if not issubclass(datatype, DataType):
            raise TypeError(f"Datatype ({datatype}) must be a subclass of {DataType}")

    def __iter__(self) -> ty.Iterable[DataType]:
        "Iterator over all items in the column, requires none of the cells to be empty"
        return (cell.item for cell in self.cells(allow_empty=False))

    def __getitem__(self, id) -> DataType:
        # TODO: could be nice to expand this to be a slice of ids
        return self.cell(id, allow_empty=False).item

    def __len__(self) -> int:
        return len(list(self.dataset.rows(self.row_frequency)))

    def cell(self, id, allow_empty: bool = True) -> DataCell:
        return DataCell.intersection(
            self,
            self.dataset.row(id=id, frequency=self.row_frequency),
            allow_empty=allow_empty,
        )

    def cells(self, allow_empty: bool = None) -> ty.Iterable[DataCell]:
        """Return an iterator over all cells in the column.

        Parameters
        ----------
        allow_empty : bool, optional
            whether to allow cells to be empty (i.e. if they don't match to an entry
            in the corresponding dataset row). If None, then cells of sink columns can
            be empty (i.e. not derived yet) and source columns can't, by default None

        Returns
        -------
        cells : Iterable[DataCell]
            an iterator over all cells in the column
        """
        return (
            DataCell.intersection(self, row, allow_empty=allow_empty)
            for row in self.dataset.rows(self.row_frequency)
        )

    @property
    def ids(self) -> list[str]:
        return [n.id for n in self.dataset.rows(self.row_frequency)]

    @property
    def path(self) -> str:
        return self._path

    def match_entry(self, row: DataRow, allow_none: bool = False) -> DataEntry:
        """Matches a single entry from a data row against the selection criteria
        defined in the column.

        Parameters
        ----------
        row: DataRow
            the row to match the item from
        allow_none: bool
            whether to return None if there are not matches

        Returns
        -------
        DataType or None
            the data item that matches the criteria/path

        Raises
        ------
        ArcanaDataMatchError
            if none or multiple items match the criteria/path of the column
            within the row
        """
        matches = row.entries
        self._mismatch_log = []
        for method in self.criteria():
            filtered = [m for m in matches if method(m)]
            if not filtered:
                if allow_none:
                    return None
                else:
                    msg = (
                        "Did not find any entries "
                        + method.__doc__.format(self=self)
                        + self._error_msg(row, matches)
                        + "\n\nDetails\n"
                        + "-------\n"
                        + "\n".join(
                            frmt.format(*ags) for frmt, ags in self._mismatch_log
                        )
                    )
                    raise ArcanaDataMatchError(msg)
            matches = filtered
        self._mismatch_log = None
        return self.select_entry_from_matches(row, matches)

    @abstractmethod
    def criteria(self) -> list[ty.Callable]:
        """returns all methods used to filter out potential matches"""

    @abstractmethod
    def format_criteria(self) -> str:
        """Formats the criteria used to match entries for use in informative error messages"""

    def matches_path(self, entry: DataEntry) -> bool:
        "that matched the path '{self.path}'"
        path, dataset_name = DataEntry.split_dataset_name_from_path(self.path)
        path_parts = self.path_split_re.split(path)
        entry_parts = self.path_split_re.split(entry.base_path)[: len(path_parts)]
        if entry_parts == path_parts:
            if dataset_name == entry.dataset_name or dataset_name == "*":
                return True
            else:
                return self._log_mismatch(
                    entry,
                    "dataset_name '{}' does not match requested '{}'",
                    entry.dataset_name,
                    dataset_name,
                )
        else:
            return self._log_mismatch(
                entry,
                "path sections {} do not match {}",
                entry_parts,
                path_parts,
            )

    def matches_datatype(self, entry: DataEntry) -> bool:
        "that matched the datatype '{self.datatype.mime_like}'"
        if self.datatype is entry.datatype:
            return True
        if not self.datatype.issubtype(entry.datatype):
            return self._log_mismatch(
                entry,
                "required datatype '{}' is not a " "sub-type of '{}'",
                self.datatype.mime_like,
                entry.datatype,
            )
        try:
            entry.get_item(self.datatype)
        except FormatMismatchError as e:
            return self._log_mismatch(entry, "datatype does not match, {}", str(e))
        else:
            return True

    def select_entry_from_matches(
        self, row: DataRow, matches: list[DataEntry]
    ) -> DataEntry:
        if len(matches) > 1:
            raise ArcanaDataMatchError(
                "Found multiple matches " + self._error_msg(row, matches)
            )
        return matches[0]

    def _error_msg(self, row: DataRow, matches: list[DataEntry]) -> str:
        row_str = f"'{row.id}' {row.frequency}" if row.id is not None else "root"
        return (
            f", when attempting to match an entry to the '{self.name}' column "
            f"in the {row_str} row of {self.dataset}\n\n  Found:"
            + self._format_matches(matches)
            + self.format_criteria()
        )

    def _format_matches(self, matches: list[DataEntry]) -> str:
        out_str = ""
        for match in sorted(matches, key=attrgetter("path")):
            out_str += "\n    "
            if match.order:
                out_str += match.order + ": "
            out_str += match.path
            out_str += f" ({match.quality})"
        return out_str

    def _log_mismatch(self, entry, format_str, *args):
        self._mismatch_log.append(("Entry {}: " + format_str, (entry.path,) + args))
        return False

    # Split a path into sections delimited by '/' or '.'
    path_split_re = re.compile(r"/|\.|@")


@attrs.define(kw_only=True)
class DataSource(DataColumn):
    """
    Specifies the criteria by which an item is selected from a data row to
    be a data source.

    Parameters
    ----------
    name: str
        the name of the column
    datatype : type
        the data type of items in the column
    row_frequency : DataSpace
        the frequency of the "rows" (data nodes) within the dataset tree, e.g. for the
        ``Clinical`` data spce the row frequency can be per 'session', 'subject',
        'timepoint', 'group', 'dataset', et...
    dataset: Dataset
        the dataset the column belongs to
    path : str
        A regex name_path to match the fileset names with. Must match
        one and only one fileset per <row_frequency>. If None, the name
        is used instead.
    quality_threshold : DataQuality
        The acceptable quality (or above) that should be considered. Data items
        will be considered missing
    order : int | None
        To be used to distinguish multiple filesets that match the
        name_path in the same session. The order of the fileset within the
        session (0-indexed). Based on the scan ID but is more robust to small
        changes to the IDs within the session if for example there are
        two scans of the same type taken before and after a task.
    required_metadata : dict[str, ty.Any]
        Required metadata, which can be used to distinguish multiple items that match all
        other criteria. The provided dictionary contains metadata values that must match
        the stored required_metadata exactly.
    is_regex : bool
        Flags whether the name_path is a regular expression or not
    """

    quality_threshold: DataQuality = attrs.field(
        default=None, converter=optional(lambda q: DataQuality[str(q)])
    )
    order: int = attrs.field(
        default=None, converter=lambda x: int(x) if x is not None else None
    )
    required_metadata: dict[str, ty.Any] = attrs.field(default=None)
    is_regex: bool = attrs.field(
        default=False,
        converter=lambda x: x.lower() == "true" if isinstance(x, str) else x,
    )

    def criteria(self) -> list[ty.Callable]:
        criteria = []
        if self.path is not None:
            if self.is_regex:
                criteria.append(self.matches_path_regex)
            else:
                criteria.append(self.matches_path)
        if self.quality_threshold is not None:
            criteria.append(self.matches_quality)
        if self.required_metadata is not None:
            criteria.append(self.matches_metadata)
        criteria.append(self.matches_datatype)
        return criteria

    def format_criteria(self) -> str:
        msg = "\n\n  Criteria: "
        if self.path:
            msg += f"\n    path='{self.path}'"
            if self.is_regex:
                msg += " (regular-expression)" if self.is_regex else ""
        if self.quality_threshold:
            msg += f"\n    quality_threshold='{self.quality_threshold}'"
        if self.required_metadata:
            msg += f"\n    required_metadata={self.required_metadata}"
        msg += f"\n    datatype='{self.datatype.mime_like}'"
        if self.order:
            msg += f"\n    order={self.order}"
        return msg

    def matches_path_regex(self, entry: DataEntry) -> bool:
        "that matched the path pattern '{self.path}'"
        pattern = self.path
        if not pattern.endswith("$"):
            # Allow paths to match with additional text after a '/' or a '.' but not
            # additional characters otherwise
            pattern += r"(?:(?:/|\.).*)?$"
        if re.match(pattern, entry.path):
            return True
        else:
            return self._log_mismatch(
                entry, "entry path {} doesn't match regular expression", pattern
            )

    def matches_quality(self, entry: DataEntry) -> bool:
        "with an acceptable quality '{self.quality_threshold}'"
        if entry.quality >= self.quality_threshold:
            return True
        else:
            return self._log_mismatch(
                entry, "quality is below threshold {}", self.quality_threshold
            )

    def matches_metadata(self, entry: DataEntry) -> bool:
        "with the required metadata '{self.required_metadata}'"
        entry_metadata = {k: entry.item_metadata[k] for k in self.required_metadata}
        if entry_metadata == self.required_metadata:
            return True
        else:
            return self._log_mismatch(
                entry,
                "metadata {} doesn't match required {}",
                entry_metadata,
                self.required_metadata,
            )

    def select_entry_from_matches(
        self, row: DataRow, matches: list[DataEntry]
    ) -> DataEntry:
        # Select a single item from the ones that match the criteria
        if self.order is not None:
            try:
                return matches[self.order - 1]
            except IndexError as e:
                raise ArcanaDataMatchError(
                    f"Not enough matching items in row {row.id} {row.frequency} in the "
                    f"'{self.name}' column to select one at index {self.order} "
                    "(starting from 1), found:" + self._format_matches(matches)
                ) from e
        else:
            return super().select_entry_from_matches(row, matches)


@attrs.define(kw_only=True)
class DataSink(DataColumn):
    """
    A specification for a file set within a analysis to be derived from a
    processing pipeline.

    Parameters
    ----------
    name: str
        the name of the column
    datatype : type
        the data type of items in the column
    row_frequency : DataSpace
        the frequency of the "rows" (data nodes) within the dataset tree, e.g. for the
        ``Clinical`` data spce the row frequency can be per 'session', 'subject',
        'timepoint', 'group', 'dataset', et...
    dataset: Dataset
        the dataset the column belongs to
    path : str
        A regex name_path to match the fileset names with. Must match
        one and only one fileset per <row_frequency>. If None, the name
        is used instead.
    salience : Salience
        The salience of the specified file-set, i.e. whether it would be
        typically of interest for publication outputs or whether it is just
        a temporary file in a workflow, and stages in between
    pipeline_name : str
        The name of the workflow applied to the dataset to generates the data
        for the sink
    """

    path = attrs.field(
        validator=attrs.validators.instance_of(str)
    )  # i.e. make mandatory
    salience: ColumnSalience = attrs.field(
        default=ColumnSalience.supplementary,
        converter=lambda s: ColumnSalience[str(s)] if s is not None else None,
    )
    pipeline_name: str = attrs.field(default=None)

    is_sink = True

    @path.default
    def path_default(self):
        return f"{self.name}@{self.dataset.name}"

    def derive(self, ids: list[str] = None):
        self.dataset.derive(self.name, ids=ids)

    def criteria(self):
        return [self.matches_path, self.matches_datatype]

    def format_criteria(self):
        return (
            "\n\n  Criteria: "
            f"\n    path='{self.path}' "
            f"\n    datatype='{self.datatype.mime_like}' "
        )

    def __setitem__(self, id, value: DataType):
        self.cell(id, allow_empty=True).item = value



# import traceback
# import pytest
# from functools import partial
# from multiprocessing import Pool, cpu_count
# from arcana.core.data.store import DataStore
# from arcana.testing.data.blueprint import TestDatasetBlueprint
# from arcana.testing import MockRemote
# from arcana.core.utils.misc import add_exc_note
from pathlib import Path
import pytest
from arcana.core.data.space import Clinical
from arcana.dirtree import DirTree
from arcana.testing.data.blueprint import TEST_DATASET_BLUEPRINTS


@pytest.mark.parametrize("blueprint_name", TEST_DATASET_BLUEPRINTS)
def test_blueprint_translation(blueprint_name: str, work_dir: Path):
    blueprint = TEST_DATASET_BLUEPRINTS[blueprint_name]
    translated = blueprint.translate_to(Clinical)
    translated.make_dataset(store=DirTree(), dataset_id=work_dir / "blueprint")


# @pytest.mark.skipif(
#     condition=cpu_count() < 2, reason="Not enough cpus to run test with multiprocessing"
# )
# def make_or_access_dataset(
#     store_name: str,
#     dataset_id: str,
#     blueprint: TestDatasetBlueprint,
#     method: str,
# ):
#     try:
#         data_store = DataStore.load(store_name)
#         if method == "make":
#             dataset = blueprint.make_dataset(data_store, dataset_id)
#         elif method == "access":
#             dataset = blueprint.access_dataset(data_store, dataset_id)
#         else:
#             assert False
#         row = next(iter(dataset.rows()))
#         return [e.path for e in row.entries]
#     except Exception as e:
#         add_exc_note(e, f"attempting to '{method}' the dataset")
#         return traceback.format_exc()


# def test_blueprint_access_dataset(
#     simple_dataset_blueprint: TestDatasetBlueprint,
#     delayed_mock_remote: MockRemote,
# ):

#     dataset_id = "blueprint_access"
#     worker = partial(
#         make_or_access_dataset,
#         delayed_mock_remote.name,
#         dataset_id,
#         simple_dataset_blueprint,
#     )
#     with Pool(2) as p:
#         try:
#             access_paths, make_paths = p.map(worker, ["access", "make"])
#         finally:
#             p.close()  # Marks the pool as closed.
#             p.join()  # Required to get the concurrency to show up in test coverage

#     assert sorted(access_paths) == ["file1", "file2"]
#     assert sorted(make_paths) == ["file1", "file2"]


# def test_blueprint_access_fail(
#     simple_dataset_blueprint: TestDatasetBlueprint, delayed_mock_remote: MockRemote
# ):
#     dataset_id = "blueprint_access_fail"
#     with pytest.raises(RuntimeError, match=f"Could not access {dataset_id}"):
#         simple_dataset_blueprint.access_dataset(
#             delayed_mock_remote, dataset_id, max_num_attempts=10, attempt_interval=0.01
#         )
